"""Test that component specification and validation works as expected using pydantic. Also test component
MISC evaluation with sparse grids.
"""
import itertools
import time
import uuid
from dataclasses import dataclass, field
from pathlib import Path

import numpy as np
from matplotlib import pyplot as plt
from numpy.typing import ArrayLike
from uqtils import approx_hess, approx_jac, ax_default

from amisc import YamlLoader
from amisc.component import Component, IndexSet, MiscTree, ModelKwargs, StringKwargs
from amisc.compression import SVD
from amisc.interpolator import Interpolator, InterpolatorState, LagrangeState
from amisc.serialize import Base64Serializable, Serializable, StringSerializable
from amisc.training import TrainingData
from amisc.typing import LATENT_STR_ID
from amisc.utils import relative_error, to_model_dataset
from amisc.variable import Variable


def test_model_kwargs():
    cases = [{'a': 1, 'b': 2, 'c': 3}, {'a': 1.1, 'b': -1.3e6, 'c': True},
             {'hello': False, 'False': (2, 2), '1': '2'},
             {'x': None, 'y': [1, 2, 3], 'z': {'key': 'value'}},
             {'pi': 3.14, 'e': 2.718, 'phi': 1.618},
             {'flag1': True, 'flag2': False, 'flag3': None},
             {'nested': {'level1': {'level2': {'level3': 'deep_value', 'level4': [True, False, (1, 2, 3)]}}}}]
    for case in cases:
        kwargs = ModelKwargs(**case)
        str_kwargs = ModelKwargs.from_dict({'method': 'string_kwargs', **case})
        serialized_kwargs = kwargs.serialize()
        str_serial = str_kwargs.serialize()
        deserialized_kwargs = ModelKwargs.deserialize(serialized_kwargs)
        str_deserialized = StringKwargs.deserialize(str_serial)
        assert deserialized_kwargs.data == kwargs.data
        assert str_deserialized.data == kwargs.data


def test_indexset():
    # Test IndexSet data structure
    Ik = [(np.random.randint(0, 10, 2), np.random.randint(0, 5, 4)) for i in range(10)]
    new_indices = [((), np.random.rand(3)*10) for i in range(5)]
    Ik = IndexSet(Ik)
    Ik.update(new_indices)
    Ik.add(((), (1.1, 2.34)))
    Ik.remove(((), (1, 2)))
    Ik.add(((1,), (-1.1, 2.3)))
    Ik = Ik.union({((1,), (1, 2, 3))})
    for alpha, beta in Ik:
        assert np.all([isinstance(ele, int) for ele in itertools.chain(alpha, beta)])
    Ik = IndexSet([str(([int(j) for j in np.random.randint(-100, 100, 7)],
                        [int(j) for j in -np.random.rand(2)*3])) for i in range(3)])
    for alpha, beta in Ik:
        assert np.all([isinstance(ele, int) for ele in itertools.chain(alpha, beta)])

    assert IndexSet.deserialize(Ik.serialize()) == Ik


def test_misctree():
    # Test MiscTree data structure
    data = {tuple([int(k) for k in np.random.randint(0, 3, 2)]):
                {tuple([int(k) for k in -np.random.rand(3)*10 + 5]): np.random.rand() for i in range(5)}
            for j in range(2)}
    new_data = {'()': {f'({i},)': np.random.rand() for i in range(4)}}
    tree = MiscTree(data)
    tree[(), (0,)] = 100
    assert tree.get(('()', '(0,)')) == 100
    tree.update(new_data)
    for alpha, beta, data in tree:
        assert np.all([isinstance(ele, int) for ele in itertools.chain(alpha, beta)])

    # Test serialization of InterpolatorState
    data = {(j, 3): {(i,): LagrangeState(weights={'x': np.random.rand(3)}, x_grids={'x': np.random.rand(3)})
                     for i in range(4)} for j in range(2)}
    tree = MiscTree(data)
    assert MiscTree.deserialize(tree.serialize()) == tree


def simple_model(x, model_fidelity=(1,), error=0.1, output_path='.'):
    return model_fidelity[0] * error + x ** 2


def test_component_validation(tmp_path):
    # Test basic component init
    def my_model(x):
        return x
    inputs = [Variable() for i in range(3)]
    outputs = [Variable() for i in range(3)]
    comp = Component(my_model, inputs, outputs)
    assert len(comp.active_set) + len(comp.candidate_set) == 0
    assert comp.xdim + comp.ydim == len(inputs) + len(outputs)
    for attribute in ['model_kwargs', 'misc_states', 'misc_costs', 'misc_coeff_train', 'misc_coeff_test',
                      'interpolator', 'training_data']:
        assert isinstance(getattr(comp, attribute), Serializable)

    # Test partial validation
    variables = [Variable(distribution=f'U(0, {np.random.rand()})') for i in range(10)]
    alphas = [(), (0,), (1,)]
    betas = [(0, 0), (0., 1.1), (1, 0)]

    active_set = []
    misc_states = {}
    misc_costs = {}
    for alpha in alphas:
        for beta in betas:
            active_set.append((alpha, beta))
            misc_costs.setdefault(alpha, dict())
            misc_states.setdefault(alpha, dict())
            misc_costs[alpha][beta] = 1.0 if np.random.rand() < 0.5 else -1.0
            misc_states[alpha][beta] = LagrangeState(weights={'x': np.random.rand(3)}, x_grids={'x': np.random.rand(3)})

    c = Component(simple_model, variables[:5], outputs=variables[5:],
                  model_kwargs={'error': 0.2}, model_fidelity=(2,), data_fidelity=(3, 2), surrogate_fidelity=(3, 3),
                  active_set=active_set, misc_states=misc_states, misc_costs=misc_costs)
    serialize_kwargs = {'training_data': {'save_path': tmp_path / 'training_data.pkl'}}
    c2 = c.serialize(serialize_kwargs=serialize_kwargs)
    c3 = Component.deserialize(c2)
    assert c3 == c


class CustomKwargs(StringKwargs):
    def __init__(self, number, **kwargs):
        self.number = number
        super().__init__(**kwargs)

    def __str__(self):
        return f'CustomKwargs({", ".join([str(self.number)] + [f"{key}={value}" for key, value in self.data.items()])})'


@dataclass
class CustomInterpolator(Interpolator, Base64Serializable):
    kernel_width: float = 0.1
    kernel_type: str = 'gaussian'

    def refine(self, *args):
        pass

    def predict(self, *args):
        pass

    def gradient(self, *args):
        pass

    def hessian(self, *args):
        pass


@dataclass
class CustomInterpolatorState(InterpolatorState, StringSerializable):
    knots: list[ArrayLike, ...] = field(default_factory=list)


@dataclass
class CustomDataStorage(TrainingData, StringSerializable):
    length: int = 40
    width: int = 2

    def get(self, *args):
        pass

    def set(self, *args):
        pass

    def refine(self, *args):
        pass

    def set_errors(self, *args):
        pass

    def impute_missing_data(self, *args):
        pass

    def clear(self):
        pass


def test_custom_data_classes():
    # Test custom args/kwargs, states, interpolator, etc.
    c = Component(simple_model, Variable(), Variable(), interpolator=CustomInterpolator(), data_fidelity=(2,),
                  training_data=CustomDataStorage(), model_kwargs=CustomKwargs(100, hello=2),
                  misc_states={(): {(): CustomInterpolatorState()}})
    c2 = c.serialize()
    c3 = Component.deserialize(c2)
    assert c3 == c


def test_save_and_load(tmp_path):
    # Test loading from yaml
    alphas = [(), (0,), (1,)]
    betas = [(0, ), (0, 0), (0, 1)]

    active_set = []
    misc_states = {}
    misc_costs = {}
    for alpha in alphas:
        for beta in betas:
            active_set.append((alpha, beta))
            misc_costs.setdefault(alpha, dict())
            misc_states.setdefault(alpha, dict())
            misc_costs[alpha][beta] = 1.0 if np.random.rand() < 0.5 else -1.0
            misc_states[alpha][beta] = LagrangeState(weights={'x': np.random.rand(5)}, x_grids={'x': np.random.rand(5)})
    c = Component(simple_model, [Variable() for i in range(3)], [Variable() for i in range(3)],
                  active_set=active_set, misc_costs=misc_costs, misc_states=misc_states,
                  model_kwargs={'output_path': '.', 'opts': {'max_iter': 1000}}, model_fidelity=(3,))

    savepath = tmp_path / 'component.yml'
    YamlLoader.dump(c, savepath)
    c_load = YamlLoader.load(savepath)
    assert c_load == c


@dataclass
class Extra:
    x1: float


def special_model(x, name='hello', frac=0.2, model_fidelity=(1,), output_path='.', output_vars=None):
    t1 = time.time()
    ret_dict = {'y1': x['x1'] + x['x2'], 'y2': x['x2'] * x['x3']}
    if np.random.rand() < frac:
        raise ValueError(f'Random error: {uuid.uuid4().hex}')
    mult = 1 if name == 'hello' else 2.5
    err = mult * (1 / model_fidelity[0])
    if output_vars is not None:
        err += output_vars['y1'].get_nominal()
    ret_dict['y3'] = np.ones((15, 2)) + err
    filename = f'output_{uuid.uuid4().hex}.txt'
    with open(Path(output_path) / filename, 'w') as fd:
        np.savetxt(fd, np.random.rand(10, 3))
    ret_dict['model_cost'] = np.random.rand() + (time.time() - t1)
    ret_dict['output_path'] = filename
    ret_dict['extra_ret'] = Extra(x1=x['x1'])
    return ret_dict


def test_model_wrapper(tmp_path):
    # Vectorized call
    def vectorized_model(x):
        return {
            'y1': x['x1'] + x['x2'],
            'y2': x['x2'] * x['x3'],
            'y3': np.ones((15, 2)) * x['x1'][..., np.newaxis, np.newaxis]
        }

    inputs = [Variable('x1'), Variable('x2'), Variable('x3')]
    outputs = [Variable('y1', nominal=0.1), Variable('y2'), Variable('y3')]
    comp = Component(vectorized_model, inputs, outputs, vectorized=True)

    cases = [{'x1': np.random.rand(10), 'x2': np.random.rand(10), 'x3': np.random.rand(10)},
             {'x1': np.random.rand(3, 5), 'x2': np.random.rand(3, 5), 'x3': np.random.rand(3, 5)},
             {'x1': np.random.rand(), 'x2': np.random.rand(), 'x3': np.random.rand()},
             [np.random.rand() for i in range(3)]]
    expected = [vectorized_model(cases[i]) for i in range(2)]
    expected += [vectorized_model({key: np.atleast_1d(cases[2][key]) for key in ['x1', 'x2', 'x3']})]
    expected += [vectorized_model({key: np.atleast_1d(cases[3][i]) for i, key in enumerate(['x1', 'x2', 'x3'])})]
    for i, case in enumerate(cases):
        y = comp.call_model(case)
        for key, val in y.items():
            assert np.allclose(val, expected[i][key])

    # Serial call
    def serial_model(x):
        return {
            'y1': float(x['x1'] + x['x2']),
            'y2': float(x['x2'] * x['x3']),
            'y3': np.ones((15, 2)) * x['x1']
        }
    comp = Component(serial_model, inputs, outputs)
    for i, case in enumerate(cases):
        y = comp.call_model(case)
        for key, val in y.items():
            assert np.allclose(val, expected[i][key])

    # Parallel call
    from concurrent.futures import ProcessPoolExecutor, ThreadPoolExecutor
    with ThreadPoolExecutor(max_workers=4) as executor:
        for i, case in enumerate(cases):
            y = comp.call_model(case, executor=executor)
            for key, val in y.items():
                assert np.allclose(val, expected[i][key])

    # Extra features (alpha, output_path, vars, exceptions, etc.)
    with ProcessPoolExecutor(max_workers=4) as executor:
        comp = Component(special_model, inputs, outputs, model_fidelity=(2,))
        shape = (5, 2, 2)
        case = {'x1': np.random.rand(*shape), 'x2': np.random.rand(*shape), 'x3': np.random.rand(*shape)}
        ret = comp.call_model(case, model_fidelity='best', output_path=tmp_path, executor=executor)
    assert ret['y3'].shape == shape + (15, 2)
    for key in ['y1', 'y2', 'extra_ret', 'model_cost', 'output_path']:
        assert ret[key].shape == shape
    y1_ravel = np.ravel(ret['y1'])
    y2_ravel = np.ravel(ret['y2'])
    y3_ravel = np.reshape(ret['y3'], (-1, 15, 2))
    if ret.get('errors'):
        for idx, error in ret.get('errors').items():
            assert 'Random error' in error['error']
            assert np.isnan(y1_ravel[idx])
            assert np.isnan(y2_ravel[idx])
            assert np.all(np.isnan(y3_ravel[idx, ...]))


def test_misc_coeff():
    """Test the iterative calculation of MISC coefficients"""
    def model(x, model_fidelity=(0, 0)):
        y = x
        return y
    comp = Component(model, model_fidelity=(2, 3), data_fidelity=(2, 1), surrogate_fidelity=(1,))
    max_ind = comp.model_fidelity + comp.max_beta

    for idx in itertools.product(*[range(m) for m in max_ind]):
        # Activate the next index
        alpha, beta = idx[:len(comp.model_fidelity)], idx[len(comp.model_fidelity):]
        neighbors = comp._neighbors(alpha, beta, forward=True)
        s = {(alpha, beta)}
        comp.update_misc_coeff(IndexSet(s), index_set='train')
        if (alpha, beta) in comp.candidate_set:
            comp.candidate_set.remove((alpha, beta))
        else:
            # Only for initial index which didn't come from the candidate set
            comp.update_misc_coeff(IndexSet(s), index_set='test')
        comp.active_set.update(s)

        comp.update_misc_coeff(neighbors, index_set='test')
        comp.candidate_set.update(neighbors)

        # Check all data structures are consistent
        coeff_sum = 0
        for a, b, coeff in comp.misc_coeff_train:
            assert (a, b) in comp.active_set and (a, b) not in comp.candidate_set
            coeff_sum += coeff
        assert coeff_sum == 1

        coeff_sum = 0
        for a, b, coeff in comp.misc_coeff_test:
            assert (a, b) in comp.active_set.union(comp.candidate_set)
            coeff_sum += coeff
        assert coeff_sum == 1


def test_sparse_grid(plots=False):
    """Simple cos test from Jakeman (2022)"""
    def model(inputs, model_fidelity=(0,)):
        alpha = np.atleast_1d(model_fidelity)
        eps = (1/5) * 2.0**(-alpha[..., 0])
        y = np.cos(np.pi/2 * (inputs['x'] + 4/5 + eps))
        return {'y': y}

    # Construct MISC surrogate from an index set
    Ik = [((0,), (0,)), ((0,), (1,)), ((1,), (0,)), ((2,), (0,)), ((1,), (1,)), ((0,), (2,)), ((1,), (2,)),
          ((2,), (1,)), ((2,), (2,))]
    x = Variable(distribution='U(-1, 1)')
    y = Variable()
    truth_alpha = (15,)
    comp = Component(model, x, y, model_fidelity=(2,), data_fidelity=(2,), vectorized=True)
    for alpha, beta in Ik:
        comp.activate_index(alpha, beta)
    N = 100
    xg = np.linspace(-1, 1, N)
    yt = comp.predict({'x': xg}, use_model=truth_alpha)['y']
    y_surr = comp.predict({'x': xg})['y']
    l2_error = relative_error(y_surr, yt)
    assert l2_error < 0.1

    # Plot results for each fidelity of the MISC surrogate
    if plots:
        fig, axs = plt.subplots(3, 3, sharey='row', sharex='col')
        for alpha in range(3):
            for beta in range(3):
                ax = axs[2-alpha, beta]
                xi, yi = comp.training_data.get((alpha,), (beta,))
                y_interp = comp.interpolator.predict({'x': xg}, comp.misc_states[(alpha,), (beta,)], (xi, yi))
                s = rf'$\hat{{f}}_{{{alpha}, {beta}}}$'
                ax.plot(xg, y_interp['y'], '--k', label=r'{}'.format(s), linewidth=1.5)
                s = rf'$\hat{{f}}_{alpha}$'
                ax.plot(xg, model({'x': xg}, (alpha,))['y'], '--b', label=r'{}'.format(s), linewidth=2)
                ax.plot(xg, yt, '-r', label=r'$f$', linewidth=2)
                ax.plot(xi['x'], yi['y'], 'or')
                ax_default(ax, r'$x$' if alpha == 0 else '', r'$f(x)$' if beta == 0 else '', legend=True)

        fig.text(0.5, 0.02, r'Increasing surrogate fidelity ($\beta$) $\rightarrow$', ha='center', fontweight='bold')
        fig.text(0.02, 0.5, r'Increasing model fidelity ($\alpha$) $\rightarrow$', va='center', fontweight='bold',
                 rotation='vertical')
        fig.set_size_inches(3 * 3, 3 * 3)
        fig.tight_layout(pad=3, w_pad=1, h_pad=1)
        plt.show()

        fig, ax = plt.subplots()
        ax.plot(xg, yt, '-r', linewidth=2, label='Model')
        ax.plot(xg, y_surr, '--k', linewidth=1.5, label='MISC surrogate')
        ax_default(ax, r'$x$', r'$f(x)$', legend=True)
        plt.show()


def test_field_quantity():
    """Test approximation of a field quantity using MISC. Both input and output."""
    def model(inputs, pressure_coords=None):
        if pressure_coords is None:
            pressure_coords = np.linspace(-1, 1, 200)
        delta = np.atleast_1d(inputs['delta'])                      # Scalar ~ U(0, 1)
        pressure_x = np.atleast_1d(inputs['pressure_x'])            # Field (N,)
        pressure_y = np.atleast_1d(inputs['pressure_y'])            # Field (N,)
        vel_x = delta[..., np.newaxis] * np.cos(pressure_coords)    # Field (N,)
        vel_y = delta[..., np.newaxis] * np.sin(pressure_coords)    # Field (N,)
        amp = delta * np.mean(pressure_x + pressure_y, axis=-1)     # Scalar

        return {'amp': amp, 'vel_x': vel_x, 'vel_y': vel_y}

    dof = 200
    num_samples = 50
    delta = Variable(distribution='U(0, 1)')
    gamma = Variable(distribution='U(1, 2)')

    # Compute compression maps from field quantity data
    grid = np.linspace(-1, 1, dof)
    delta_samples = delta.sample(num_samples)
    gamma_samples = gamma.sample(num_samples)
    pressure_x = delta_samples[..., np.newaxis] * np.sin(grid)
    pressure_y = gamma_samples[..., np.newaxis] * np.cos(grid)
    pressure_matrix = {'pressure_x': pressure_x, 'pressure_y': pressure_y}
    pressure = Variable(compression=SVD(rank=2, data_matrix=pressure_matrix, coords=grid,
                                        fields=['pressure_x', 'pressure_y']), name='pressure')
    ret = model({'delta': delta_samples, 'pressure_x': pressure_x, 'pressure_y': pressure_y})
    vel_x, vel_y = ret['vel_x'], ret['vel_y']
    vel_matrix = {'vel_x': vel_x, 'vel_y': vel_y}
    vel = Variable(compression=SVD(rank=4, data_matrix=vel_matrix, coords=grid, fields=['vel_x', 'vel_y']))

    amp = Variable()

    max_beta = (3, 3)
    comp = Component(model, [delta, pressure], [amp, vel], data_fidelity=max_beta, vectorized=True)
    for idx in itertools.product(*[range(beta+1) for beta in max_beta]):
        comp.activate_index((), idx)

    xtest = {'delta': delta.sample(10)}
    pressure_samples = pressure.sample_domain(10)  # (10, rank)
    xtest.update({f'pressure{LATENT_STR_ID}{i}': pressure_samples[:, i] for i in range(pressure_samples.shape[1])})
    xtest_model = to_model_dataset(xtest, comp.inputs)[0]

    ysurr = comp.predict(xtest)
    ysurr = to_model_dataset(ysurr, comp.outputs)[0]
    ymodel = comp.predict(xtest_model, use_model='best', pressure_coords=grid)
    errors = {var: relative_error(ysurr[var], ymodel[var]) for var in ymodel}
    assert all([errors[var] < 0.01 for var in errors])


def test_comp_jacobian_and_hessian():
    f1 = lambda theta: 2 * theta['x1'] ** 2 * theta['x2'] + np.cos(theta['x3'])
    f2 = lambda theta: np.exp(theta['x2']) * theta['x1'] + np.sin(theta['x3']) * theta['x2']
    fun = lambda theta: {'y1': f1(theta), 'y2': f2(theta)}

    x1, x2, x3 = (Variable('x1', distribution='U(0, 2)'), Variable('x2', distribution='U(-1, 1)'),
                  Variable('x3', distribution='U(-3.14, 3.14)'))
    y1, y2 = Variable('y1'), Variable('y2')
    max_beta = (4, 4, 5)

    surr = Component(fun, [x1, x2, x3], [y1, y2], data_fidelity=max_beta, vectorized=True)
    for idx in itertools.product(*[range(beta) for beta in max_beta]):
        surr.activate_index((), idx)

    N = (5, 6)

    def fun_vec(theta: np.ndarray):
        theta_dict = {var: theta[..., i] for i, var in enumerate(['x1', 'x2', 'x3'])}
        y_dict = fun(theta_dict)
        y_vec = np.concatenate([y_dict[var][..., np.newaxis] for var in ['y1', 'y2']], axis=-1)
        return y_vec

    xtest = {'x1': x1.sample(N), 'x2': x2.sample(N), 'x3': x3.sample(N)}
    xvec = np.concatenate([xtest[var][..., np.newaxis] for var in ['x1', 'x2', 'x3']], axis=-1)

    jac_truth = approx_jac(fun_vec, xvec)
    jac_surr = surr.gradient(xtest, derivative='first')
    jac_surr_vec = np.concatenate([np.expand_dims(jac_surr[var], axis=-2) for var in ['y1', 'y2']], axis=-2)
    assert np.allclose(jac_truth, jac_surr_vec, rtol=1e-2, atol=1e-2)

    hess_truth = approx_hess(fun_vec, xvec)
    hess_surr = surr.hessian(xtest)
    hess_surr_vec = np.concatenate([np.expand_dims(hess_surr[var], axis=-3) for var in ['y1', 'y2']], axis=-3)
    assert np.allclose(hess_truth, hess_surr_vec, rtol=1e-1, atol=1e-1)


def test_get_training_data():
    def model(inputs, model_fidelity=(0,)):
        return {'y': inputs['x'] ** 2 + model_fidelity[0]}

    x = Variable(domain=(0, 3))
    y = Variable()
    comp = Component(model, x, y, model_fidelity=(3,), data_fidelity=(3,))
    for idx in itertools.product(*[range(alpha+1) for alpha in comp.model_fidelity],
                                 *[range(beta+1) for beta in comp.data_fidelity]):
        a = idx[:len(comp.model_fidelity)]
        b = idx[len(comp.model_fidelity):]
        comp.activate_index(a, b)

    xtrain, ytrain = comp.get_training_data()
    xtrue, ytrue = comp.training_data.get(comp.model_fidelity, comp.max_beta)
    assert all([np.allclose(xtrain[var], xtrue[var]) for var in xtrue])
    assert all([np.allclose(ytrain[var], ytrue[var]) for var in ytrue])



"""Provides interpolator classes. Interpolators approximate the input &rarr; output mapping of a model given
a set of training data. The training data consists of input-output pairs, and the interpolator can be
refined with new training data.

Includes:

- `Interpolator`: Abstract class providing basic structure of an interpolator
- `Lagrange`: Concrete implementation for tensor-product barycentric Lagrange interpolation
- `InterpolatorState`: Interface for a dataclass that stores the internal state of an interpolator
- `LagrangeState`: The internal state for a barycentric Lagrange polynomial interpolator
"""
from __future__ import annotations

import copy
import itertools
from abc import ABC, abstractmethod
from dataclasses import dataclass, field

import numpy as np

from amisc.serialize import Base64Serializable, Serializable, StringSerializable
from amisc.typing import Dataset, MultiIndex

__all__ = ["InterpolatorState", "LagrangeState", "Interpolator", "Lagrange"]


class InterpolatorState(Serializable, ABC):
    """Interface for a dataclass that stores the internal state of an interpolator (e.g. weights and biases)."""
    pass


@dataclass
class LagrangeState(InterpolatorState, Base64Serializable):
    """The internal state for a barycentric Lagrange polynomial interpolator.

    :ivar weights: the 1d interpolation grid weights
    :ivar x_grids: the 1d interpolation grids
    """
    weights: dict[str, np.ndarray] = field(default_factory=dict)
    x_grids: dict[str, np.ndarray] = field(default_factory=dict)

    def __eq__(self, other):
        if isinstance(other, LagrangeState):
            try:
                return all([np.allclose(self.weights[var], other.weights[var]) for var in self.weights]) and \
                    all([np.allclose(self.x_grids[var], other.x_grids[var]) for var in self.x_grids])
            except IndexError:
                return False
        else:
            return False


class Interpolator(Serializable, ABC):
    """Interface for an interpolator object that approximates a model. An interpolator should:

    - `refine` - take an old state and new training data and produce a new "refined" state (e.g. new weights/biases)
    - `predict` - interpolate from the training data to a new set of points (i.e. approximate the underlying model)
    - `gradient` - compute the grdient/Jacobian at new points (if you want)
    - `hessian` - compute the 2nd derivative/Hessian at new points (if you want)

    Currently, only the `Lagrange` interpolator is supported and can be constructed from a configuration `dict`
    via `Interpolator.from_dict()`.
    """

    @abstractmethod
    def refine(self, beta: MultiIndex, training_data: tuple[Dataset, Dataset],
               old_state: InterpolatorState, input_domains: dict[str, tuple]) -> InterpolatorState:
        """Refine the interpolator state with new training data.

        :param beta: a multi-index specifying the fidelity "levels" of the new interpolator state (starts at (0,... 0))
        :param training_data: a tuple of `xi, yi` Datasets for the input/output training data
        :param old_state: the previous state of the interpolator (None if initializing the first state)
        :param input_domains: a `dict` mapping input variables to their corresponding domains
        :returns: the new "refined" interpolator state
        """
        raise NotImplementedError

    @abstractmethod
    def predict(self, x: dict | Dataset, state: InterpolatorState, training_data: tuple[Dataset, Dataset]) -> Dataset:
        """Interpolate the output of the model at points `x` using the given state and training data

        :param x: the input Dataset `dict` mapping input variables to locations at which to compute the interpolator
        :param state: the current state of the interpolator
        :param training_data: a tuple of `xi, yi` Datasets for the input/output training data for the current state
        :returns: a Dataset `dict` mapping output variables to interpolator outputs
        """
        raise NotImplementedError

    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)

    @abstractmethod
    def gradient(self, x: Dataset, state: InterpolatorState, training_data: tuple[Dataset, Dataset]) -> Dataset:
        """Evaluate the gradient/Jacobian at points `x` using the interpolator.

        :param x: the input Dataset `dict` mapping input variables to locations at which to evaluate the Jacobian
        :param state: the current state of the interpolator
        :param training_data: a tuple of `xi, yi` Datasets for the input/output training data for the current state
        :returns: a Dataset `dict` mapping output variables to Jacobian matrices of shape `(ydim, xdim)` -- for
                  scalar outputs, the Jacobian is returned as `(xdim,)`
        """
        raise NotImplementedError

    @abstractmethod
    def hessian(self, x: Dataset, state: InterpolatorState, training_data: tuple[Dataset, Dataset]) -> Dataset:
        """Evaluate the Hessian at points `x` using the interpolator.

        :param x: the input Dataset `dict` mapping input variables to locations at which to evaluate the Hessian
        :param state: the current state of the interpolator
        :param training_data: a tuple of `xi, yi` Datasets for the input/output training data for the current state
        :returns: a Dataset `dict` mapping output variables to Hessian matrices of shape `(xdim, xdim)`
        """
        raise NotImplementedError

    @classmethod
    def from_dict(cls, config: dict) -> Interpolator:
        """Create an `Interpolator` object from a `dict` config. Only `method='lagrange'` is supported for now."""
        method = config.pop('method', 'lagrange').lower()
        match method:
            case 'lagrange':
                return Lagrange(**config)
            case other:
                raise NotImplementedError(f"Unknown interpolator method: {other}")


@dataclass
class Lagrange(Interpolator, StringSerializable):
    """Implementation of a tensor-product barycentric Lagrange polynomial interpolator. A `LagrangeState` stores
    the 1d interpolation grids and weights for each input dimension. `Lagrange` computes the tensor-product
    of 1d Lagrange polynomials to approximate a multi-variate function.

    :ivar interval_capacity: tuning knob for Lagrange interpolation (see Berrut and Trefethen 2004)
    """
    interval_capacity: float = 4.0

    @staticmethod
    def _extend_grids(x_grids: dict[str, np.ndarray], x_points: dict[str, np.ndarray]):
        """Extend the 1d `x` grids with any new points from `x_points`, skipping duplicates. This will preserve the
        order of new points in the extended grid without duplication. This will maintain the same order as
        `SparseGrid.x_grids` if that is the underlying training data structure.

        !!! Example
            ```python
            x = {'x': np.array([0, 1, 2])}
            new_x = {'x': np.array([3, 0, 2, 3, 1, 4])}
            extended_x = Lagrange._extend_grids(x, new_x)
            # gives {'x': np.array([0, 1, 2, 3, 4])}
            ```

        !!! Warning
            This will only work for 1d grids; all `x_grids` should be scalar quantities. Field quantities should
            already be passed in as several separate 1d latent coefficients.

        :param x_grids: the current 1d interpolation grids
        :param x_points: the new points to extend the interpolation grids with
        :returns: the extended grids
        """
        extended_grids = copy.deepcopy(x_grids)
        for var, new_pts in x_points.items():
            # Get unique new values that are not already in the grid (maintain their order; keep only first one)
            u = x_grids[var] if var in x_grids else np.array([])
            u, ind = np.unique(new_pts[~np.isin(new_pts, u)], return_index=True)
            u = u[np.argsort(ind)]
            extended_grids[var] = u if var not in x_grids else np.concatenate((x_grids[var], u), axis=0)

        return extended_grids

    def refine(self, beta: MultiIndex, training_data: tuple[Dataset, Dataset],
               old_state: LagrangeState, input_domains: dict[str, tuple]) -> LagrangeState:
        """Refine the interpolator state with new training data.

        :param beta: the refinement level indices for the interpolator (not used for `Lagrange`)
        :param training_data: a tuple of dictionaries containing the new training data (`xtrain`, `ytrain`)
        :param old_state: the old interpolator state to refine (None if initializing)
        :param input_domains: a `dict` of each input variable's domain; input keys should match `xtrain` keys
        :returns: the new interpolator state
        """
        xtrain, ytrain = training_data  # Lagrange only really needs the xtrain data to update barycentric weights/grids

        # Initialize the interpolator state
        if old_state is None:
            x_grids = self._extend_grids({}, xtrain)
            weights = {}
            for var, grid in x_grids.items():
                bds = input_domains[var]
                Nx = grid.shape[0]
                C = (bds[1] - bds[0]) / self.interval_capacity  # Interval capacity (see Berrut and Trefethen 2004)
                xj = grid.reshape((Nx, 1))
                xi = xj.reshape((1, Nx))
                dist = (xj - xi) / C
                np.fill_diagonal(dist, 1)  # Ignore product when i==j
                weights[var] = (1.0 / np.prod(dist, axis=1))  # (Nx,)

        # Otherwise, refine the interpolator state
        else:
            x_grids = self._extend_grids(old_state.x_grids, xtrain)
            weights = copy.deepcopy(old_state.weights)
            for var, grid in x_grids.items():
                bds = input_domains[var]
                Nx_old = old_state.x_grids[var].shape[0]
                Nx_new = grid.shape[0]
                if Nx_new > Nx_old:
                    weights[var] = np.pad(weights[var], [(0, Nx_new - Nx_old)], mode='constant', constant_values=np.nan)
                    C = (bds[1] - bds[0]) / self.interval_capacity
                    for j in range(Nx_old, Nx_new):
                        weights[var][:j] *= (C / (grid[:j] - grid[j]))
                        weights[var][j] = np.prod(C / (grid[j] - grid[:j]))

        return LagrangeState(weights=weights, x_grids=x_grids)

    def predict(self, x: Dataset, state: LagrangeState, training_data: tuple[Dataset, Dataset]):
        """Predict the output of the model at points `x` with barycentric Lagrange interpolation."""
        # Convert `x` and `yi` to 2d arrays: (N, xdim) and (N, ydim)
        # Inputs `x` may come in unordered, but they should get realigned with the internal `x_grids` state
        xi, yi = training_data
        x_arr = np.concatenate([x[var][..., np.newaxis] for var in xi], axis=-1)
        yi_arr = np.concatenate([yi[var][..., np.newaxis] for var in yi], axis=-1)

        xdim = x_arr.shape[-1]
        ydim = yi_arr.shape[-1]
        grid_sizes = {var: grid.shape[-1] for var, grid in state.x_grids.items()}
        max_size = max(grid_sizes.values())
        dims = list(range(xdim))

        # Create ragged edge matrix of interpolation pts and weights
        x_j = np.full((xdim, max_size), np.nan)                             # For example:
        w_j = np.full((xdim, max_size), np.nan)                             # A= [#####--
        for n, var in enumerate(state.x_grids):                             #     #######
            x_j[n, :grid_sizes[var]] = state.x_grids[var]                   #     ###----]
            w_j[n, :grid_sizes[var]] = state.weights[var]

        diff = x_arr[..., np.newaxis] - x_j
        div_zero_idx = np.isclose(diff, 0, rtol=1e-4, atol=1e-8)
        check_interp_pts = np.sum(div_zero_idx) > 0     # whether we are evaluating directly on some interp pts
        diff[div_zero_idx] = 1
        quotient = w_j / diff                           # (..., xdim, Nx)
        qsum = np.nansum(quotient, axis=-1)             # (..., xdim)
        y = np.zeros(x_arr.shape[:-1] + (ydim,))        # (..., ydim)

        # Loop over multi-indices and compute tensor-product lagrange polynomials
        indices = [range(s) for s in grid_sizes.values()]
        for i, j in enumerate(itertools.product(*indices)):
            L_j = quotient[..., dims, j] / qsum         # (..., xdim)

            # Set L_j(x==x_j)=1 for the current j and set L_j(x==x_j)=0 for x_j = x_i, i != j
            if check_interp_pts:
                other_pts = np.copy(div_zero_idx)
                other_pts[div_zero_idx[..., dims, j]] = False
                L_j[div_zero_idx[..., dims, j]] = 1
                L_j[np.any(other_pts, axis=-1)] = 0

            # Add multivariate basis polynomial contribution to interpolation output
            y += np.prod(L_j, axis=-1, keepdims=True) * yi_arr[i, :]

        # Unpack the outputs back into a Dataset
        y_ret = {}
        start_idx = 0
        for var, arr in yi.items():
            num_vals = arr.shape[-1] if len(arr.shape) > 1 else 1
            end_idx = start_idx + num_vals
            y_ret[var] = y[..., start_idx:end_idx]
            if len(arr.shape) == 1:
                y_ret[var] = np.squeeze(y_ret[var], axis=-1)  # for scalars
            start_idx = end_idx
        return y_ret

    def gradient(self, x: Dataset, state: LagrangeState, training_data: tuple[Dataset, Dataset]):
        """Evaluate the gradient/Jacobian at points `x` using the interpolator."""
        # Convert `x` and `yi` to 2d arrays: (N, xdim) and (N, ydim)
        xi, yi = training_data
        x_arr = np.concatenate([x[var][..., np.newaxis] for var in xi], axis=-1)
        yi_arr = np.concatenate([yi[var][..., np.newaxis] for var in yi], axis=-1)

        xdim = x_arr.shape[-1]
        ydim = yi_arr.shape[-1]
        grid_sizes = {var: grid.shape[-1] for var, grid in state.x_grids.items()}
        max_size = max(grid_sizes.values())

        # Create ragged edge matrix of interpolation pts and weights
        x_j = np.full((xdim, max_size), np.nan)                             # For example:
        w_j = np.full((xdim, max_size), np.nan)                             # A= [#####--
        for n, var in enumerate(state.x_grids):                             #     #######
            x_j[n, :grid_sizes[var]] = state.x_grids[var]                   #     ###----]
            w_j[n, :grid_sizes[var]] = state.weights[var]

        # Compute values ahead of time that will be needed for the gradient
        diff = x_arr[..., np.newaxis] - x_j
        div_zero_idx = np.isclose(diff, 0, rtol=1e-4, atol=1e-8)
        check_interp_pts = np.sum(div_zero_idx) > 0
        diff[div_zero_idx] = 1
        quotient = w_j / diff                               # (..., xdim, Nx)
        qsum = np.nansum(quotient, axis=-1)                 # (..., xdim)
        sqsum = np.nansum(w_j / diff ** 2, axis=-1)         # (..., xdim)
        jac = np.zeros(x_arr.shape[:-1] + (ydim, xdim))     # (..., ydim, xdim)

        # Loop over multi-indices and compute derivative of tensor-product lagrange polynomials
        indices = [range(s) for s in grid_sizes.values()]
        for k, var in enumerate(grid_sizes):
            dims = [idx for idx in range(xdim) if idx != k]
            for i, j in enumerate(itertools.product(*indices)):
                j_dims = [j[idx] for idx in dims]
                L_j = quotient[..., dims, j_dims] / qsum[..., dims]  # (..., xdim-1)

                # Partial derivative of L_j with respect to x_k
                dLJ_dx = ((w_j[k, j[k]] / (qsum[..., k] * diff[..., k, j[k]])) *
                          (sqsum[..., k] / qsum[..., k] - 1 / diff[..., k, j[k]]))

                # Set L_j(x==x_j)=1 for the current j and set L_j(x==x_j)=0 for x_j = x_i, i != j
                if check_interp_pts:
                    other_pts = np.copy(div_zero_idx)
                    other_pts[div_zero_idx[..., list(range(xdim)), j]] = False
                    L_j[div_zero_idx[..., dims, j_dims]] = 1
                    L_j[np.any(other_pts[..., dims, :], axis=-1)] = 0

                    # Set derivatives when x is at the interpolation points (i.e. x==x_j)
                    p_idx = [idx for idx in range(grid_sizes[var]) if idx != j[k]]
                    w_j_large = np.broadcast_to(w_j[k, :], x_arr.shape[:-1] + w_j.shape[-1:]).copy()
                    curr_j_idx = div_zero_idx[..., k, j[k]]
                    other_j_idx = np.any(other_pts[..., k, :], axis=-1)
                    dLJ_dx[curr_j_idx] = -np.nansum((w_j[k, p_idx] / w_j[k, j[k]]) /
                                                    (x_arr[curr_j_idx, k, np.newaxis] - x_j[k, p_idx]), axis=-1)
                    dLJ_dx[other_j_idx] = ((w_j[k, j[k]] / w_j_large[other_pts[..., k, :]]) /
                                           (x_arr[other_j_idx, k] - x_j[k, j[k]]))

                dLJ_dx = np.expand_dims(dLJ_dx, axis=-1) * np.prod(L_j, axis=-1, keepdims=True)  # (..., 1)

                # Add contribution to the Jacobian
                jac[..., k] += dLJ_dx * yi_arr[i, :]

        # Unpack the outputs back into a Dataset (array of length xdim for each y_var giving partial derivatives)
        jac_ret = {}
        start_idx = 0
        for var, arr in yi.items():
            num_vals = arr.shape[-1] if len(arr.shape) > 1 else 1
            end_idx = start_idx + num_vals
            jac_ret[var] = jac[..., start_idx:end_idx, :]  # (..., ydim, xdim)
            if len(arr.shape) == 1:
                jac_ret[var] = np.squeeze(jac_ret[var], axis=-2)  # for scalars: (..., xdim) partial derivatives
            start_idx = end_idx
        return jac_ret

    def hessian(self, x: Dataset, state: LagrangeState, training_data: tuple[Dataset, Dataset]):
        """Evaluate the Hessian at points `x` using the interpolator."""
        # Convert `x` and `yi` to 2d arrays: (N, xdim) and (N, ydim)
        xi, yi = training_data
        x_arr = np.concatenate([x[var][..., np.newaxis] for var in xi], axis=-1)
        yi_arr = np.concatenate([yi[var][..., np.newaxis] for var in yi], axis=-1)

        xdim = x_arr.shape[-1]
        ydim = yi_arr.shape[-1]
        grid_sizes = {var: grid.shape[-1] for var, grid in state.x_grids.items()}
        grid_size_list = list(grid_sizes.values())
        max_size = max(grid_size_list)

        # Create ragged edge matrix of interpolation pts and weights
        x_j = np.full((xdim, max_size), np.nan)                             # For example:
        w_j = np.full((xdim, max_size), np.nan)                             # A= [#####--
        for n, var in enumerate(state.x_grids):                             #     #######
            x_j[n, :grid_sizes[var]] = state.x_grids[var]                   #     ###----]
            w_j[n, :grid_sizes[var]] = state.weights[var]

        # Compute values ahead of time that will be needed for the gradient
        diff = x_arr[..., np.newaxis] - x_j
        div_zero_idx = np.isclose(diff, 0, rtol=1e-4, atol=1e-8)
        check_interp_pts = np.sum(div_zero_idx) > 0
        diff[div_zero_idx] = 1
        quotient = w_j / diff                                       # (..., xdim, Nx)
        qsum = np.nansum(quotient, axis=-1)                         # (..., xdim)
        qsum_p = -np.nansum(w_j / diff ** 2, axis=-1)               # (..., xdim)
        qsum_pp = 2 * np.nansum(w_j / diff ** 3, axis=-1)           # (..., xdim)

        # Loop over multi-indices and compute 2nd derivative of tensor-product lagrange polynomials
        hess = np.zeros(x_arr.shape[:-1] + (ydim, xdim, xdim))      # (..., ydim, xdim, xdim)
        indices = [range(s) for s in grid_size_list]
        for m in range(xdim):
            for n in range(m, xdim):
                dims = [idx for idx in range(xdim) if idx not in [m, n]]
                for i, j in enumerate(itertools.product(*indices)):
                    j_dims = [j[idx] for idx in dims]
                    L_j = quotient[..., dims, j_dims] / qsum[..., dims]  # (..., xdim-2)

                    # Set L_j(x==x_j)=1 for the current j and set L_j(x==x_j)=0 for x_j = x_i, i != j
                    if check_interp_pts:
                        other_pts = np.copy(div_zero_idx)
                        other_pts[div_zero_idx[..., list(range(xdim)), j]] = False
                        L_j[div_zero_idx[..., dims, j_dims]] = 1
                        L_j[np.any(other_pts[..., dims, :], axis=-1)] = 0

                    # Cross-terms in Hessian
                    if m != n:
                        # Partial derivative of L_j with respect to x_m and x_n
                        d2LJ_dx2 = np.ones(x_arr.shape[:-1])
                        for k in [m, n]:
                            dLJ_dx = ((w_j[k, j[k]] / (qsum[..., k] * diff[..., k, j[k]])) *
                                      (-qsum_p[..., k] / qsum[..., k] - 1 / diff[..., k, j[k]]))

                            # Set derivatives when x is at the interpolation points (i.e. x==x_j)
                            if check_interp_pts:
                                p_idx = [idx for idx in range(grid_size_list[k]) if idx != j[k]]
                                w_j_large = np.broadcast_to(w_j[k, :], x_arr.shape[:-1] + w_j.shape[-1:]).copy()
                                curr_j_idx = div_zero_idx[..., k, j[k]]
                                other_j_idx = np.any(other_pts[..., k, :], axis=-1)
                                dLJ_dx[curr_j_idx] = -np.nansum((w_j[k, p_idx] / w_j[k, j[k]]) /
                                                                (x_arr[curr_j_idx, k, np.newaxis] - x_j[k, p_idx]),
                                                                axis=-1)
                                dLJ_dx[other_j_idx] = ((w_j[k, j[k]] / w_j_large[other_pts[..., k, :]]) /
                                                       (x_arr[other_j_idx, k] - x_j[k, j[k]]))

                            d2LJ_dx2 *= dLJ_dx

                        d2LJ_dx2 = np.expand_dims(d2LJ_dx2, axis=-1) * np.prod(L_j, axis=-1, keepdims=True)  # (..., 1)
                        hess[..., m, n] += d2LJ_dx2 * yi_arr[i, :]
                        hess[..., n, m] += d2LJ_dx2 * yi_arr[i, :]

                    # Diagonal terms in Hessian:
                    else:
                        front_term = w_j[m, j[m]] / (qsum[..., m] * diff[..., m, j[m]])
                        first_term = (-qsum_pp[..., m] / qsum[..., m]) + 2 * (qsum_p[..., m] / qsum[..., m]) ** 2
                        second_term = (2 * (qsum_p[..., m] / (qsum[..., m] * diff[..., m, j[m]]))
                                       + 2 / diff[..., m, j[m]] ** 2)
                        d2LJ_dx2 = front_term * (first_term + second_term)

                        # Set derivatives when x is at the interpolation points (i.e. x==x_j)
                        if check_interp_pts:
                            curr_j_idx = div_zero_idx[..., m, j[m]]
                            other_j_idx = np.any(other_pts[..., m, :], axis=-1)
                            if np.any(curr_j_idx) or np.any(other_j_idx):
                                p_idx = [idx for idx in range(grid_size_list[m]) if idx != j[m]]
                                w_j_large = np.broadcast_to(w_j[m, :], x_arr.shape[:-1] + w_j.shape[-1:]).copy()
                                x_j_large = np.broadcast_to(x_j[m, :], x_arr.shape[:-1] + x_j.shape[-1:]).copy()

                                # if these points are at the current j interpolation point
                                d2LJ_dx2[curr_j_idx] = (2 * np.nansum((w_j[m, p_idx] / w_j[m, j[m]]) /
                                                                      (x_arr[curr_j_idx, m, np.newaxis] - x_j[m, p_idx]), # noqa: E501
                                                                      axis=-1) ** 2 +
                                                        2 * np.nansum((w_j[m, p_idx] / w_j[m, j[m]]) /
                                                                      (x_arr[curr_j_idx, m, np.newaxis] - x_j[m, p_idx]) ** 2, # noqa: E501
                                                                      axis=-1))

                                # if these points are at any other interpolation point
                                other_pts_inv = other_pts.copy()
                                other_pts_inv[other_j_idx, m, :grid_size_list[m]] = np.invert(
                                    other_pts[other_j_idx, m, :grid_size_list[m]])  # noqa: E501
                                curr_x_j = x_j_large[other_pts[..., m, :]].reshape((-1, 1))
                                other_x_j = x_j_large[other_pts_inv[..., m, :]].reshape((-1, len(p_idx)))
                                curr_w_j = w_j_large[other_pts[..., m, :]].reshape((-1, 1))
                                other_w_j = w_j_large[other_pts_inv[..., m, :]].reshape((-1, len(p_idx)))
                                curr_div = w_j[m, j[m]] / np.squeeze(curr_w_j, axis=-1)
                                curr_diff = np.squeeze(curr_x_j, axis=-1) - x_j[m, j[m]]
                                d2LJ_dx2[other_j_idx] = ((-2 * curr_div / curr_diff) * (np.nansum(
                                    (other_w_j / curr_w_j) / (curr_x_j - other_x_j), axis=-1) + 1 / curr_diff))

                        d2LJ_dx2 = np.expand_dims(d2LJ_dx2, axis=-1) * np.prod(L_j, axis=-1, keepdims=True)  # (..., 1)
                        hess[..., m, n] += d2LJ_dx2 * yi_arr[i, :]

        # Unpack the outputs back into a Dataset (matrix (xdim, xdim) for each y_var giving 2nd partial derivatives)
        hess_ret = {}
        start_idx = 0
        for var, arr in yi.items():
            num_vals = arr.shape[-1] if len(arr.shape) > 1 else 1
            end_idx = start_idx + num_vals
            hess_ret[var] = hess[..., start_idx:end_idx, :, :]  # (..., ydim, xdim, xdim)
            if len(arr.shape) == 1:
                hess_ret[var] = np.squeeze(hess_ret[var], axis=-3)  # for scalars: (..., xdim, xdim) partial derivatives
            start_idx = end_idx
        return hess_ret



"""Test convergence of `System.fit` to a few smooth analytical models."""
from pathlib import Path
from typing import Literal

import numpy as np

from amisc.system import System
from amisc.utils import relative_error


def converge(system: Literal['borehole', 'wing'] = 'borehole', num_samples=1000):
    """Test convergence of different model surrogates."""
    surr = System.load_from_file(Path(__file__).parent / f'{system.lower()}.yml')

    xt = surr.sample_inputs(num_samples)
    yt = surr.predict(xt, use_model='best')

    surr.fit()
    ysurr = surr.predict(xt)
    l2_error = {var: relative_error(ysurr[var], yt[var]) for var in yt}

    return l2_error


def test_borehole(tol=1e-2):
    l2_error = converge(system='borehole')
    assert np.max(l2_error['vdot']) < tol


def test_wing_weight(tol=1e-2):
    l2_error = converge(system='wing')
    assert np.max(l2_error['Wwing']) < tol



"""Test serialization mixin classes."""
from amisc.serialize import Base64Serializable, PickleSerializable, Serializable, StringSerializable, YamlSerializable


class ClassTestBase64(Base64Serializable):
    def __init__(self, value):
        self.value = value


def test_base64_serializable():
    obj = ClassTestBase64(42)
    serialized = obj.serialize()
    deserialized = ClassTestBase64.deserialize(serialized)
    assert obj.value == deserialized.value


class ClassTestString(StringSerializable):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return f'TestClass({self.value})'


def test_string_serializable():
    obj = ClassTestString(42)
    serialized = obj.serialize()
    deserialized = ClassTestString.deserialize(serialized)
    assert obj.value == deserialized.value


class ClassTestPickle(PickleSerializable):
    def __init__(self, value):
        self.value = value


def test_pickle_serializable(tmp_path):
    fname = tmp_path / 'test_pickle.pkl'
    obj = ClassTestPickle(42)
    serialized = obj.serialize(fname)  # noqa: F841
    deserialized = ClassTestPickle.deserialize(fname)
    assert obj.value == deserialized.value


class ClassTestYaml(Serializable):
    def __init__(self, value):
        self.value = value

    def serialize(self):
        return {'value': self.value}

    @classmethod
    def deserialize(cls, serialized_data):
        return cls(serialized_data['value'])


def test_yaml_serializable():
    meta_obj = YamlSerializable(obj=ClassTestYaml)
    serialized = meta_obj.serialize()
    deserialized = YamlSerializable.deserialize(serialized)
    assert meta_obj.obj == deserialized.obj



"""Tests for the `amisc` package. Ordered according to the following hierarchy of modules, with higher-level
modules dependent on the lower-levels.

Testing modules:

- `test_utils` - tests for the `amisc.utils` module including dataset conversion, formatting, and other utilities.
- `test_serialize` - tests for the `amisc.serialize` module including serialization mixin classes.
- `test_variable` - tests for the `amisc.variable` module including the `Variable` and `VariableList` classes,
                    as well as `Transform`, `Distribution`, and `Compression` interfaces.
- `test_training_data` - tests for the `amisc.training` module including the `SparseGrid` implementation.
- `test_interpolator` - tests for the `amisc.interpolator` module including the `Lagrange` implementation.
- `test_component` - tests for the `amisc.component` module including the `Component` model wrapper and its
                     associated data structures `MiscTree` and `IndexSet`.
- `test_system_spec` - tests for the `amisc.system` module, primarily focused on pydantic validation and serialization.
- `test_system` - tests for the `amisc.system` module, primarily focused on the `System` class and its methods.
- `test_convergence` - tests for the `amisc.system` module to ensure convergence of the `System` surrogate on
                       various test models.

Profiling modules:

- `profile_fit` - profile the performance of `System.fit` for various configurations.
- `profile_predict` - profile the performance of `System.predict` using `py-spy`.
"""


"""Test `System` prediction and refinement methods."""
import copy
import time
import warnings
from concurrent.futures import ThreadPoolExecutor

import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import fsolve
from scipy.stats import gaussian_kde
from uqtils import ax_default

from amisc import Component, System, Variable
from amisc.compression import SVD
from amisc.utils import relative_error, to_model_dataset


def test_feedforward_simple(plots=False):
    """Test the MD system in Figure 5 in Jakeman 2022."""
    from amisc.examples.models import f1, f2

    surr = System(f1, f2)

    x = {'x': np.linspace(0, 1, 100)}
    y1_ret = f1(x['x'])
    y2_ret = f2(f1(x['x']))
    y_truth = {'y1': y1_ret, 'y2': y2_ret}
    y_surr = surr.predict(x, use_model='best')
    l2_error = np.array([relative_error(y_surr[var], y_truth[var]) for var in ['y1', 'y2']])
    assert np.all(l2_error < 1e-15)

    if plots:
        fig, ax = plt.subplots(1, 2, figsize=(6, 3), layout='tight')
        ax[0].plot(x['x'], y_truth['y1'], '-k', label='$f_1(x)$')
        ax[0].plot(x['x'], y_surr['y1'], '--r', label='Surrogate')
        ax_default(ax[0], '$x$', '$f(x)$', legend=True)
        ax[1].plot(x['x'], y_truth['y2'], '-k', label='$f(x)$')
        ax[1].plot(x['x'], y_surr['y2'], '--r', label='Surrogate')
        ax_default(ax[1], '$x$', '$f(x)$', legend=True)
        plt.show()


def test_feedforward_three(plots=False):
    """Test the MD system in Figure 6 in Jakeman 2022."""
    def coupled_system(D1, D2, Q1, Q2):
        """Scalable number of inputs (D1, D2) and outputs (Q1, Q2)."""
        def f1(inputs, model_fidelity=None):
            if model_fidelity is None:
                model_fidelity = np.ones((1, 1)) * 15
            eps = 10. ** (-model_fidelity[..., 0])
            outputs = {}
            for q in range(1, Q1+1):
                sum = np.sum(np.concatenate([inputs[f'z1_{i}'][np.newaxis, ...] for i in range(1, D1+1)],
                                            axis=0), axis=0)
                outputs[f'y1_{q}'] = inputs['z1_1'] ** q * np.sin(sum + eps)
            return outputs

        def f2(inputs, model_fidelity=None):
            alpha = model_fidelity
            if alpha is None:
                alpha = np.ones((1, 1)) * 15
            eps = 10. ** (-alpha[..., 0])
            outputs = {}
            for q in range(1, Q2+1):
                prod1 = np.prod(np.concatenate([inputs[f'y1_{i}'][np.newaxis, ...] ** (q+1) - eps
                                                for i in range(1, Q1+1)], axis=0), axis=0)
                prod2 = np.prod(np.concatenate([inputs[f'z2_{i}'][np.newaxis, ...] for i in range(1, D2+1)],
                                               axis=0), axis=0)
                outputs[f'y2_{q}'] = prod1 * prod2
            return outputs

        def f3(inputs, model_fidelity=None, D3=D1):
            alpha = model_fidelity
            if alpha is None:
                alpha = np.ones((1, 1)) * 15
            eps = 10. ** (-alpha[..., 0])
            sum1 = np.sum(np.concatenate([(inputs[f'y2_{i}'][np.newaxis, ...] - eps) ** 2 for i in range(1, Q2+1)],
                                         axis=0), axis=0)
            sum2 = np.sum(np.concatenate([inputs[f'z1_{i}'][np.newaxis, ...] for i in range(1, D3+1)], axis=0), axis=0)
            y3 = np.exp(-sum1) / (1 + (25/16) * sum2 ** 2)
            return {'y3_1': y3}

        return f1, f2, f3

    # Hook up the 'wiring' for this example feedforward system
    D1 = 1
    D2 = D1
    Q1 = 1
    Q2 = Q1
    alpha = (15,)
    inputs = ([Variable(f'z1_{i}', distribution='U(0, 1)') for i in range(1, D1+1)] +
              [Variable(f'z2_{i}', distribution='U(0, 1)') for i in range(1, D2+1)])
    outputs = ([Variable(f'y1_{i}') for i in range(1, Q1+1)] + [Variable(f'y2_{i}') for i in range(1, Q2+1)] +
               [Variable('y3_1')])
    f1, f2, f3 = coupled_system(D1, D2, Q1, Q2)
    comp1 = Component(f1, inputs[:D1], outputs[:Q1], vectorized=True, model_fidelity=alpha)
    comp2 = Component(f2, inputs[D1:] + outputs[:Q1], outputs[Q1:Q1+Q2], vectorized=True, model_fidelity=alpha)
    comp3 = Component(f3, inputs[:D1] + outputs[Q1:Q1+Q2], outputs[Q1+Q2:], vectorized=True, model_fidelity=alpha)
    surr = System(comp1, comp2, comp3)

    # Test example
    N = 5000
    z = surr.sample_inputs(N)
    all_inputs = copy.deepcopy(z)
    y1_truth = f1(z)
    all_inputs.update(y1_truth)
    y2_truth = f2(all_inputs)
    all_inputs.update(y2_truth)
    y3_truth = f3(all_inputs)
    y1_truth.update({**y2_truth, **y3_truth})
    y_surr = surr.predict(z, use_model='best')

    for var in y1_truth:
        l2_error = relative_error(y_surr[var], y1_truth[var])
        assert np.max(l2_error) < 1e-15

    # Show coupling variable pdfs
    if plots:
        # Typo in the figure: rho(y1) is flipped and rho(y3) is not so exaggerated
        fig, ax = plt.subplots()
        ls = ['-r', '--k', ':b']
        pts = np.linspace(0, 1, 100)
        for i, var in enumerate(y1_truth):
            label_str = f'$\\rho(y_{{{i+1}}})$'
            kernel = gaussian_kde(y_surr[var])
            # ax[i].hist(y_surr[:, i], density=True, bins=20, color='r', edgecolor='black', linewidth=1.2)
            ax.plot(pts, kernel(pts), ls[i], label=label_str)
        ax.set_xlabel(r'$y$')
        ax.set_ylabel('PDF')
        ax.legend()
        fig.tight_layout()
        plt.show()


def test_fpi():
    """Test fixed point iteration implementation against scipy fsolve."""
    f1 = lambda x: {'y1': -x['x']**3 + 2 * x['y2']**2}
    f2 = lambda x: {'y2': 3*x['x']**2 + 4 * x['y1']**(-2)}
    x = Variable('x', distribution='U(1, 4)')
    y1 = Variable('y1', distribution='U(1, 10)')
    y2 = Variable('y2', distribution='U(1, 10)')
    comp1 = Component(f1, [x, y2], y1, name='m1')
    comp2 = Component(f2, [x, y1], y2, name='m2')
    surr = System(comp1, comp2)

    # Test on random x against scipy.fsolve
    N = 100
    tol = 1e-12
    x0 = np.array([5.5, 5.5])
    exo = surr.sample_inputs(N)
    y_surr = surr.predict(exo, use_model='best', anderson_mem=10, max_fpi_iter=200, fpi_tol=tol)  # (N, 2)
    nan_idx = np.full(N, False)
    for var, arr in y_surr.items():
        nan_idx = np.logical_or(nan_idx, np.isnan(arr))
    nan_idx = list(nan_idx.nonzero()[0])
    y_true = np.zeros((N, 2))
    bad_idx = []
    warnings.simplefilter('error')
    for i in range(N):
        def fun(x):
            y1 = x[0]
            y2 = x[1]
            res1 = -exo['x'][i]**3 + 2*y2**2 - y1
            res2 = 3*exo['x'][i]**2 + 4*y1**(-2) - y2
            return [res1, res2]

        try:
            y_true[i, :] = fsolve(fun, x0, xtol=tol)
        except Exception:
            bad_idx.append(i)

    y_surr = np.vstack([y_surr[var] for var in ['y1', 'y2']]).T
    y_surr = np.delete(y_surr, nan_idx + bad_idx, axis=0)
    y_true = np.delete(y_true, nan_idx + bad_idx, axis=0)
    l2_error = relative_error(y_surr, y_true)
    assert np.max(l2_error) < tol


def test_system_refine(plots=False):
    """Test iterative refinement for Figure 5 in Jakeman 2022."""
    from amisc.examples.models import f1, f2

    beta_max = (4,)
    surr = System(Component(f1, name='f1', vectorized=True, data_fidelity=beta_max),
                  Component(f2, name='f2', vectorized=True, data_fidelity=beta_max))

    for var in surr.variables():
        var.domain = (0, 1)

    Niter = 4
    x = np.linspace(0, 1, 100)
    y1 = f1(x)
    y2 = f2(x)
    y3 = f2(y1)

    # Initialize all components
    for i in range(len(surr.components)):
        surr.refine()

    fig, ax = plt.subplots(Niter, 3, sharex='col', sharey='row')
    for i in range(Niter):
        # Plot actual function values
        ax[i, 0].plot(x, y1, '-r', label='$f_1(x)$')
        ax[i, 1].plot(x, y2, '-r', label='$f_2(y)$')
        ax[i, 2].plot(x, y3, '-r', label='$f(x)$')

        # Plot first component surrogates
        comp = surr.get_component('f1')
        ax[i, 0].plot(x, comp.predict({'x': x}, index_set='train')['y1'], '--k', label='$f_1$ current')
        beta_max = 0
        for alpha, beta in comp.active_set:
            if beta[0] > beta_max:
                beta_max = beta[0]
        xi, yi = comp.get_training_data((), (beta_max,))
        ax[i, 0].plot(xi['x'], yi['y1'], 'ok', markersize=8, label='')
        for alpha, beta in comp.candidate_set:
            yJ1 = surr.predict({'x': x}, index_set={comp.name: {(alpha, beta)}}, incremental={comp.name: True})['y2']
            y1_cand = comp.predict({'x': x}, index_set={(alpha, beta)}, incremental=True)['y1']
            ax[i, 0].plot(x, y1_cand, ':b', label='$f_1$ candidate')
            xi, yi = comp.get_training_data(alpha, beta)
            ax[i, 0].plot(xi['x'], yi['y1'], 'xb', markersize=8, label='')

        # Plot second component surrogates
        comp = surr.get_component('f2')
        ax[i, 1].plot(x, comp.predict({'y1': x}, index_set='train')['y2'], '--k', label='$f_2$ current')
        beta_max = 0
        for alpha, beta in comp.active_set:
            if beta[0] > beta_max:
                beta_max = beta[0]
        xi, yi = comp.get_training_data((), (beta_max,))
        ax[i, 1].plot(xi['y1'], yi['y2'], 'ok', markersize=8, label='')
        for alpha, beta in comp.candidate_set:
            yJ2 = surr.predict({'x': x}, index_set={comp.name: {(alpha, beta)}}, incremental={comp.name: True})['y2']
            y2_cand = comp.predict({'y1': x}, index_set={(alpha, beta)}, incremental=True)['y2']
            ax[i, 1].plot(x, y2_cand, '-.g', label='$f_2$ candidate')
            xi, yi = comp.get_training_data(alpha, beta)
            ax[i, 1].plot(xi['y1'], yi['y2'], 'xg', markersize=8, label='')

        # Plot integrated surrogates
        ysurr = surr.predict({'x': x}, index_set='train')['y2']
        ax[i, 2].plot(x, ysurr, '--k', label='$f_J$')
        ax[i, 2].plot(x, yJ1, ':b', label='$f_{J_1}$')
        ax[i, 2].plot(x, yJ2, '-.g', label='$f_{J_2}$')
        ax_default(ax[i, 0], '$x$', '$f_1(x)$', legend=True)
        ax_default(ax[i, 1], '$y$', '$f_2(y)$', legend=True)
        ax_default(ax[i, 2], '$x$', '$f_2(f_1(x))$', legend=True)

        # Refine the system
        surr.refine()

    ysurr = surr.predict({'x': x})
    ytrue = surr.predict({'x': x}, use_model='best')
    for var in ysurr:
        l2_error = relative_error(ysurr[var], ytrue[var])
        assert np.max(l2_error) < 0.1

    if plots:
        fig.set_size_inches(3.5*3, 3.5*Niter)
        fig.tight_layout()
        plt.show()


def test_simulate_fit():
    """Test looping back through training history and recomputing on a test set."""
    from amisc.examples.models import f1, f2

    beta_max = (4,)
    surr = System(Component(f1, name='f1', vectorized=True, data_fidelity=beta_max),
                  Component(f2, name='f2', vectorized=True, data_fidelity=beta_max))

    for var in surr.variables():
        var.domain = (0, 1)

    xtest = {'x': np.linspace(0, 1, 100)}
    y1_ret = f1(xtest['x'])
    y2_ret = f2(f1(xtest['x']))
    ytest = {'y1': y1_ret, 'y2': y2_ret}

    max_iter = 8
    surr.fit(max_iter=max_iter, test_set=(xtest, ytest))

    i = 0
    for train_res, active_set, candidate_set, misc_coeff_train, misc_coeff_test in surr.simulate_fit():
        error = train_res['test_error']
        if all([~np.isnan(val) for val in error.values()]):
            test_index_set = {comp: active_set[comp].union(candidate_set[comp]) for comp in active_set}
            y_surr_train = surr.predict(xtest, index_set=active_set, misc_coeff=misc_coeff_train)
            y_surr_test = surr.predict(xtest, index_set=test_index_set, misc_coeff=misc_coeff_test)

            train_error = {var: relative_error(y_surr_train[var], ytest[var]) for var in ytest}
            test_error = {var: relative_error(y_surr_test[var], ytest[var]) for var in ytest}

            # Training error should be the same as was originally computed during fit
            assert all([np.allclose(error[var], train_error[var]) for var in train_error])

            # Error using the candidate set should be less than the training error
            assert all([np.all(test_error[var] <= train_error[var]) for var in test_error])

        i += 1

    assert i == max_iter


def test_fit_with_executor(model_cost=1, max_iter=5, max_workers=8):
    """Run fit with an executor."""

    def expensive_model(inputs, model_cost=1):
        time.sleep(model_cost)
        return {'y': -inputs['x1'] ** 3 + 2 * inputs['x2'] ** 2 + inputs['x3']}

    inputs = [Variable('x1', distribution='U(0, 1)'), Variable('x2', distribution='U(0, 1)'),
              Variable('x3', distribution='U(0, 1)')]
    outputs = Variable('y')
    comp = Component(expensive_model, inputs, outputs, data_fidelity=(3, 3, 3), model_cost=model_cost)
    surr = System(comp)

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        surr.fit(max_iter=max_iter, executor=executor)


def test_fire_sat(tmp_path, plots=False):
    """Test the fire satellite coupled system from Chaudhuri (2018)."""
    from amisc.examples.models import fire_sat_system

    N = 100
    surr = fire_sat_system(save_dir=tmp_path)

    targs = ['Slew', 'Asa', 'Pat']
    xt = surr.sample_inputs(N, use_pdf=True)
    yt = surr.predict(xt, use_model='best', targets=targs)
    use_idx = np.full(N, True)
    for var, arr in yt.items():
        use_idx = np.logical_and(use_idx, ~np.isnan(arr))

    xt = {var: arr[use_idx] for var, arr in xt.items()}
    yt = {var: arr[use_idx] for var, arr in yt.items()}

    # with ProcessPoolExecutor(max_workers=4) as executor:
    executor=None
    surr.fit(targets=targs, max_iter=10, max_tol=1e-3, runtime_hr=4/12, test_set=(xt, yt), executor=executor,
             plot_interval=4, estimate_bounds=True)

    ysurr = surr.predict(xt, targets=targs)
    for var in yt:
        l2_error = relative_error(ysurr[var], yt[var])
        assert np.nanmax(l2_error) < 0.2

    # Plot 1d slices
    slice_idx = ['H', 'Po', 'Cd']
    try:
        surr.plot_slice(slice_idx, targs, show_model=['best', 'worst'], model_dir=surr.root_dir,
                        random_walk=True, num_steps=15)
    except np.linalg.LinAlgError:
        print('Its alright. Sometimes the random walks are wacky and FPI wont converge.')

    # Plot allocation bar charts
    surr.plot_allocation()

    if plots:
        # Plot error histograms
        fig, ax = plt.subplots(1, 3)
        kwargs = {'bins': 20, 'edgecolor': 'black', 'density': False, 'linewidth': 1.2}
        for i, var in enumerate(targs):
            ax[i].hist(yt[var], color='r', label='Truth', **kwargs)
            ax[i].hist(ysurr[var], color='b', alpha=0.4, label='Surrogate', **kwargs)
            ax_default(ax[i], surr.outputs()[var].get_tex(symbol=False, units=True), '', legend=True)

        fig.set_size_inches(9, 3)
        fig.tight_layout()
        plt.show()


def test_turbojet_cycle(tmp_path):
    """Test for coupling a compressor, combuster, and turbine models. Includes a mix of analytical and surrogate
    models, as well as coupling field quantities along with scalar random variables.
    """
    def compressor(inputs):
        """Simple compressor model.

        :param inputs: `dict` of `Ta` and `v1` for ambient temperature and inlet velocity distribution
        :returns: `dict` of modified velocity `v2` and pressure distribution `p1`
        """
        inlet_velocity_distribution = inputs['v1']
        ambient_temp = inputs['Ta']
        compression_ratio = 1.5
        modified_velocity = inlet_velocity_distribution * compression_ratio
        modified_velocity += ambient_temp * 0.05
        base_pressure = 100
        pressure_boost_factor = 1.2
        pressure_distribution = base_pressure + (modified_velocity * pressure_boost_factor)
        return {'v2': modified_velocity, 'p1': pressure_distribution}

    def combuster(inputs, model_fidelity=None):
        """Simple combuster model.

        :param inputs: `dict` of `v2` and `Ta` for compressor outlet velocity and ambient temperature.
        :param model_fidelity: `tuple` of fidelity levels for combustion factor and temperature effect
        :returns: `dict` of outlet velocity `v3`
        """
        # Fidelity levels for combustion factor and external input effect
        alpha = model_fidelity or (3, 3)
        combustion_factor = 1.5 + 0.5 * alpha[0]
        temperature_factor = 0.05 + 0.03 * alpha[1]

        # Applying the combustion effect
        outlet_velocity = inputs['v2'] * combustion_factor + inputs['Ta'] * temperature_factor

        return {'v3': outlet_velocity}

    def turbine(inputs, model_fidelity=None):
        """Simple turbine model.

        :param inputs: `dict` of `v3` and `eta_t` for combuster outlet velocity and blade efficiency.
        :param model_fidelity: `tuple` of fidelity levels for turbine efficiency and other effects.
        :returns: `dict` of exit velocity `v4`
        """
        # Fidelity levels for turbine efficiency and other effects
        alpha = model_fidelity or (3,)
        turbine_efficiency = 0.7 + 0.1 * alpha[0]
        pressure_effect = 0.01 + 0.005 * alpha[0]
        efficiency_effect = 0.01 + 0.005 * alpha[0]

        # Applying the turbine efficiency effect
        exit_velocity_field = inputs['v3'] * turbine_efficiency

        # Adjustments based on pressure field and external input effects
        blade_efficiency = inputs['eta_t']
        exit_velocity_field += inputs['p1'] * pressure_effect
        exit_velocity_field += efficiency_effect * blade_efficiency

        return {'v4': exit_velocity_field}

    # Synthetic velocity distributions
    num_samples = 100
    num_points = 80
    inlet_grid = np.linspace(-1, 1, num_points)
    means = np.random.uniform(200, 300, num_samples)
    stds = np.random.uniform(0.15, 0.35, num_samples)
    v1_samples = np.zeros((num_points, num_samples))
    for i in range(num_samples):
        v1_samples[:, i] = means[i] * np.exp(-0.5 * (inlet_grid / stds[i]) ** 2)

    # Random variables
    Ta = Variable('Ta', distribution='N(300, 20)', units='K', description='Ambient temperature', norm='zscore')
    eta_t = Variable('eta_t', distribution='N(0.8, 0.04)', description='Turbine blade efficiency')
    Ta_samples = Ta.sample(num_samples)
    eta_t_samples = eta_t.sample(num_samples)

    # Field quantities
    v1 = Variable('v1', units='m/s', description='Inlet velocity', compression=SVD(rank=4, coords=inlet_grid))
    v2 = Variable('v2', units='m/s', description='Compressor outlet velocity',
                  compression=SVD(rank=3, coords=inlet_grid))
    v3 = Variable('v3', units='m/s', description='Combuster outlet velocity',
                  compression=SVD(rank=2, coords=inlet_grid))
    v4 = Variable('v4', units='m/s', description='Turbine exit velocity',
                  compression=SVD(rank=4, coords=inlet_grid))
    p1 = Variable('p1', units='kPa', description='Compressor outlet pressure',
                  compression=SVD(rank=2, coords=inlet_grid))

    # Define components
    comp1 = Component(compressor, [Ta, v1], [v2, p1], name='Compressor')
    comp2 = Component(combuster, [Ta, v2], [v3], data_fidelity=(2, 2), model_fidelity=(3, 3), name='Combuster')
    comp3 = Component(turbine, [eta_t, v3, p1], [v4], data_fidelity=(2, 2, 2), model_fidelity=(3,), name='Turbine')
    system = System(comp1, comp2, comp3, root_dir=tmp_path)

    # Generate compression data
    xtest = {'Ta': Ta_samples, 'v1': v1_samples.T, 'eta_t': eta_t_samples}
    ytest = system.predict(xtest, use_model='best', normalized_inputs=False)
    v1.compression.compute_map(data_matrix=v1_samples)
    v2.compression.compute_map(data_matrix=ytest['v2'].T)
    v3.compression.compute_map(data_matrix=ytest['v3'].T)
    v4.compression.compute_map(data_matrix=ytest['v4'].T)
    p1.compression.compute_map(data_matrix=ytest['p1'].T)

    # Set domain estimate for v1 latent coefficients
    v1_latent = v1.compression.compress(v1_samples.T)
    v1.update_domain(list(zip(np.min(v1_latent, axis=0), np.max(v1_latent, axis=0))), override=True)

    system.fit(max_iter=15, test_set=(xtest, ytest), estimate_bounds=True)

    # Check error
    ysurr_norm = system.predict(xtest, normalized_inputs=False)
    ysurr = to_model_dataset(ysurr_norm, system.outputs())[0]
    for var in ysurr:
        assert relative_error(ysurr[var], ytest[var]) < 0.15



import logging
import re
from pathlib import Path
from urllib.parse import quote

import coverage
from coverage import Coverage

PROJECT_DIR = Path(__file__).parent.parent
PYTEST_STATUS_FILE = PROJECT_DIR / '.pytest_status'
README_FILE = PROJECT_DIR / 'README.md'
COVERAGE_FILE = PROJECT_DIR / '.coverage'


def update_readme_coverage_badge():
    """Update the test coverage (codecov) badge in the README with most recent pytest-cov results."""
    # Get total coverage percentage from .coverage file
    try:
        cov = Coverage(data_file=COVERAGE_FILE)
        cov.load()
        total = round(cov.report())
    except coverage.exceptions.NoDataError:
        print('No coverage data found. Skipping README update...')
        return

    # Get badge color
    color_bds = [(95, "brightgreen"), (90, "green"), (75, "yellowgreen"), (60, "yellow"), (40, "orange"), (0, "red")]
    badge_color = 'lightgrey'
    try:
        xtotal = int(total)
    except ValueError:
        pass
    else:
        for low_bound, color in color_bds:
            if xtotal >= low_bound:
                badge_color = color
                break
    badge_url = f"https://img.shields.io/badge/coverage-{total}{quote('%')}-{badge_color}?logo=codecov"

    empty_badge_pattern = "![Code Coverage]()"
    existing_badge_pattern = r"\!\[Code Coverage\]\(.*?\)"
    patterns_to_match = rf"{re.escape(empty_badge_pattern)}|{existing_badge_pattern}"
    replacement_str = f"![Code Coverage]({badge_url})"

    # Use regex to replace badge icon in README
    with open(README_FILE, "r+", encoding='utf-8') as f:
        text = f.read()
        if replacement_str in text:
            logging.info('Coverage in README is up to date -- no changes made.')
        elif bool(re.search(patterns_to_match, text)) is False:
            logging.warning(f"Couldn't find the pattern {empty_badge_pattern} in README.md.")
        else:
            updated_text = re.sub(patterns_to_match, replacement_str, text)
            f.seek(0)
            f.write(updated_text)
            f.truncate()
            logging.info("README has been successfully updated with test coverage.")


def pytest_sessionfinish(session, exitstatus):
    """Do things after pytest session finishes."""
    logging.shutdown()
    update_readme_coverage_badge()

    # Write an exitstatus to file (for pre-commit checking)
    with open(PYTEST_STATUS_FILE, 'w') as f:
        f.write(str(exitstatus))



"""Module for probability distribution functions (PDFs).

Includes:

- `Distribution` — an abstract interface for specifying a PDF.
- `Uniform` — a uniform distribution.
- `Normal` — a normal distribution.
- `Relative` — a relative distribution (i.e. uniform within a percentage of a nominal value).
- `Tolerance` — a tolerance distribution (i.e. uniform within a tolerance of a nominal value).
- `LogUniform` — a log-uniform distribution.
- `LogNormal` — a log-normal distribution.

Distribution objects can be converted easily to/from strings for serialization.
"""
from __future__ import annotations

from abc import ABC, abstractmethod

import numpy as np

from amisc.utils import parse_function_string

__all__ = ['Distribution', 'Uniform', 'Normal', 'Relative', 'Tolerance', 'LogNormal', 'LogUniform']


class Distribution(ABC):
    """Base class for PDF distributions that provide sample and pdf methods. Distributions should:

    - `sample` - return samples from the distribution
    - `pdf` - return the probability density function of the distribution

    :ivar dist_args: the arguments that define the distribution (e.g. mean and std for a normal distribution)
    :vartype dist_args: tuple
    """

    def __init__(self, dist_args: tuple):
        self.dist_args = dist_args

    def __str__(self):
        """Serialize a `Distribution` object to/from string."""
        return f'{type(self).__name__}{self.dist_args}'

    def __repr__(self):
        return self.__str__()

    def domain(self, dist_args: tuple = None) -> tuple:
        """Return the domain of this distribution. Defaults to `dist_args`

        :param dist_args: overrides `self.dist_args`
        """
        return dist_args or self.dist_args

    def nominal(self, dist_args: tuple = None) -> float:
        """Return the nominal value of this distribution. Defaults to middle of domain.

        :param dist_args: overrides `self.dist_args`
        """
        lb, ub = self.domain(dist_args=dist_args)
        return (lb + ub) / 2

    @classmethod
    def from_string(cls, dist_string: str) -> Distribution | None:
        """Convert a string to a `Distribution` object.

        :param dist_string: specifies a PDF or distribution. Can be `Normal(mu, std)`, `Uniform(lb, ub)`,
                            `LogUniform(lb, ub)`, `LogNormal(mu, std)`,
                            `Relative(pct)`, or `Tolerance(tol)`. The shorthands `N(0, 1)`, `U(0, 1)`, `LU(0, 1)`,
                            `LN(0, 1)`, `rel(5)`, or `tol(1)` are also accepted.
        :return: the corresponding `Distribution` object
        """
        if not dist_string:
            return None

        dist_name, args, kwargs = parse_function_string(dist_string)
        if dist_name in ['N', 'Normal', 'normal']:
            # Normal distribution like N(0, 1)
            try:
                mu = float(kwargs.get('mu', args[0]))
                std = float(kwargs.get('std', args[1]))
                return Normal((mu, std))
            except Exception as e:
                raise ValueError(f'Normal distribution string "{dist_string}" is not valid: Try N(0, 1).') from e
        elif dist_name in ['U', 'Uniform', 'uniform']:
            # Uniform distribution like U(0, 1)
            try:
                lb = float(kwargs.get('lb', args[0]))
                ub = float(kwargs.get('ub', args[1]))
                return Uniform((lb, ub))
            except Exception as e:
                raise ValueError(f'Uniform distribution string "{dist_string}" is not valid: Try U(0, 1).') from e
        elif dist_name in ['R', 'Relative', 'relative', 'rel']:
            # Relative uniform distribution like rel(+-5%)
            try:
                pct = float(kwargs.get('pct', args[0]))
                return Relative((pct,))
            except Exception as e:
                raise ValueError(f'Relative distribution string "{dist_string}" is not valid: Try rel(5).') from e
        elif dist_name in ['T', 'Tolerance', 'tolerance', 'tol']:
            # Uniform distribution within a tolerance like tol(+-1)
            try:
                tol = float(kwargs.get('tol', args[0]))
                return Tolerance((tol,))
            except Exception as e:
                raise ValueError(f'Tolerance distribution string "{dist_string}" is not valid: Try tol(1).') from e
        elif dist_name in ['LogUniform', 'LU']:
            # LogUniform distribution like LU(1e-3, 1e-1)
            try:
                lb = float(kwargs.get('lb', args[0]))
                ub = float(kwargs.get('ub', args[1]))
                base = float(kwargs.get('base', args[2] if len(args) > 2 else 10))
                return LogUniform((lb, ub), base=base)
            except Exception as e:
                raise ValueError(f'LogUniform distr string "{dist_string}" is not valid: Try LU(1e-3, 1e-1).') from e
        elif dist_name in ['LogNormal', 'LN']:
            # LogNniform distribution like LN(-2, 1)
            try:
                mu = float(kwargs.get('mu', args[0]))
                std = float(kwargs.get('std', args[1]))
                base = float(kwargs.get('base', args[2] if len(args) > 2 else 10))
                return LogNormal((mu, std), base=base)
            except Exception as e:
                raise ValueError(f'LogNormal distr string "{dist_string}" is not valid: Try LN(-2, 1).') from e
        else:
            raise NotImplementedError(f'The distribution "{dist_string}" is not recognized.')

    @abstractmethod
    def sample(self, shape: int | tuple, nominal: float | np.ndarray = None, dist_args: tuple = None) -> np.ndarray:
        """Sample from the distribution.

        :param shape: shape of the samples to return
        :param nominal: a nominal value(s) for sampling (e.g. for relative distributions)
        :param dist_args: overrides `Distribution.dist_args`
        :return: the samples of the given shape
        """
        raise NotImplementedError

    @abstractmethod
    def pdf(self, x: np.ndarray, dist_args: tuple = None) -> np.ndarray:
        """Evaluate the pdf of this distribution at the `x` locations.

        :param x: the locations at which to evaluate the pdf
        :param dist_args: overrides `Distribution.dist_args`
        :return: the pdf evaluations
        """
        raise NotImplementedError


class Uniform(Distribution):
    """A Uniform distribution. Specify by string as "Uniform(lb, ub)" or "U(lb, ub)" in shorthand."""

    def __str__(self):
        return f'U({self.dist_args[0]}, {self.dist_args[1]})'

    def sample(self, shape, nominal=None, dist_args=None):
        lb, ub = dist_args or self.dist_args
        return np.random.rand(*shape) * (ub - lb) + lb

    def pdf(self, x, dist_args=None):
        x = np.atleast_1d(x)
        lb, ub = dist_args or self.dist_args
        pdf = np.broadcast_to(1 / (ub - lb), x.shape).copy()
        pdf[np.where(x > ub)] = 0
        pdf[np.where(x < lb)] = 0
        return pdf


class LogUniform(Distribution):
    """A LogUniform distribution. Specify by string as `LogUniform(lb, ub)` or `LU(lb, ub)` in shorthand. Uses
    base-10 by default.

    !!! Example
        ```python
        x = LogUniform((1e-3, 1e-1))  # log10(x) ~ U(-3, -1)
        ```
    """
    def __init__(self, dist_args: tuple, base=10):
        self.base = base
        super().__init__(dist_args)

    def __str__(self):
        return f'LU({self.dist_args[0]}, {self.dist_args[1]}, base={self.base})'

    def sample(self, shape, nominal=None, dist_args=None):
        lb, ub = dist_args or self.dist_args
        c = 1 / np.log(self.base)
        return self.base ** (np.random.rand(*shape) * c * (np.log(ub) - np.log(lb)) + c * np.log(lb))

    def pdf(self, x, dist_args=None):
        x = np.atleast_1d(x)
        lb, ub = dist_args or self.dist_args
        c = 1 / np.log(self.base)
        const = 1 / (c * (np.log(ub) - np.log(lb)) * np.log(self.base))
        pdf = const / x
        pdf[np.where(x > ub)] = 0
        pdf[np.where(x < lb)] = 0
        return pdf


class Normal(Distribution):
    """A Normal distribution. Specify by string as `Normal(mu, std)` or `N(mu, std)` in shorthand."""

    def __str__(self):
        return f'N({self.dist_args[0]}, {self.dist_args[1]})'

    def domain(self, dist_args=None):
        """Defaults the domain of the distribution to 3 standard deviations above and below the mean."""
        mu, std = dist_args or self.dist_args
        return mu - 3 * std, mu + 3 * std

    def sample(self, shape, nominal=None, dist_args=None):
        mu, std = dist_args or self.dist_args
        return np.random.randn(*shape) * std + mu

    def pdf(self, x, dist_args=None):
        mu, std = dist_args or self.dist_args
        return (1 / (np.sqrt(2 * np.pi) * std)) * np.exp(-0.5 * ((x - mu) / std) ** 2)


class LogNormal(Distribution):
    """A LogNormal distribution. Specify by string as `LogNormal(mu, sigma)` or `LN(mu, sigma)` in shorthand.
    Uses base-10 by default.

    !!! Example
        ```python
        x = LogNormal((-2, 1))  # log10(x) ~ N(-2, 1)
        ```
    """
    def __init__(self, dist_args: tuple, base=10):
        self.base = base
        super().__init__(dist_args)

    def __str__(self):
        return f'LN({self.dist_args[0]}, {self.dist_args[1]}, base={self.base})'

    def domain(self, dist_args=None):
        """Defaults the domain of the distribution to 3 standard deviations above and below the mean."""
        mu, std = dist_args or self.dist_args
        return self.base ** (mu - 3 * std), self.base ** (mu + 3 * std)

    def sample(self, shape, nominal=None, dist_args=None):
        mu, std = dist_args or self.dist_args
        return self.base ** (np.random.randn(*shape) * std + mu)

    def pdf(self, x, dist_args=None):
        mu, std = dist_args or self.dist_args
        const = (1 / (np.sqrt(2 * np.pi) * std * x * np.log(self.base)))
        return const * np.exp(-0.5 * ((np.log(x)/np.log(self.base) - mu) / std) ** 2)


class Relative(Distribution):
    """A Relative distribution. Specify by string as `Relative(pct)` or `rel(pct%)` in shorthand.
    Will attempt to sample uniformly within the given percent of a nominal value.
    """

    def __str__(self):
        return rf'rel({self.dist_args[0]})'

    def domain(self, dist_args=None):
        return None

    def nominal(self, dist_args=None):
        return None

    def sample(self, shape, nominal=None, dist_args=None):
        if nominal is None:
            raise ValueError('Cannot sample relative distribution when no nominal value is provided.')
        dist_args = dist_args or self.dist_args
        tol = abs((dist_args[0] / 100) * nominal)
        return np.random.rand(*shape) * 2 * tol - tol + nominal

    def pdf(self, x, dist_args=None):
        return np.ones(x.shape)


class Tolerance(Distribution):
    """A Tolerance distribution. Specify by string as `Tolerance(tol)` or `tol(tol)` in shorthand.
    Will attempt to sample uniformly within a given absolute tolerance of a nominal value.
    """

    def __str__(self):
        return rf'tol({self.dist_args[0]})'

    def domain(self, dist_args=None):
        return None

    def nominal(self, dist_args=None):
        return None

    def sample(self, shape, nominal=None, dist_args=None):
        if nominal is None:
            raise ValueError('Cannot sample tolerance distribution when no nominal value is provided.')
        dist_args = dist_args or self.dist_args
        tol = abs(dist_args[0])
        return np.random.rand(*shape) * 2 * tol - tol + nominal

    def pdf(self, x, dist_args=None):
        return np.ones(np.atleast_1d(x).shape)



"""Provides an object-oriented interface for model inputs/outputs, random variables, scalars, and field quantities.

Includes:

- `Variable` — an object that stores information about a variable and includes methods for sampling, pdf evaluation,
               normalization, compression, loading from file, etc. Variables can mostly be treated as strings
               that have some additional information and utilities attached to them.
- `VariableList` — a container for `Variables` that provides dict-like access of `Variables` by `name` along with normal
                   indexing and slicing.

The preferred serialization of `Variable` and `VariableList` is to/from yaml. This is done by default with the
`!Variable` and `!VariableList` yaml tags.
"""
from __future__ import annotations

import ast
import random
import string
from collections import OrderedDict
from pathlib import Path
from typing import ClassVar, Optional, Union

import numpy as np
import yaml
from numpy.typing import ArrayLike
from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator

from amisc.compression import Compression
from amisc.distribution import Distribution, LogUniform, Normal, Uniform
from amisc.serialize import Serializable
from amisc.transform import Minmax, Transform, Zscore
from amisc.typing import LATENT_STR_ID, CompressionData
from amisc.utils import _get_yaml_path, _inspect_assignment, search_for_file

__all__ = ['Variable', 'VariableList']
_TransformLike = Union[str, Transform, list[str | Transform]]  # something that can be converted to a Transform


class Variable(BaseModel, Serializable):
    """Object for storing information about variables and providing methods for pdf evaluation, sampling, etc.
    All fields will undergo pydantic validation and conversion to the correct types.

    A simple variable object can be created with `var = Variable()`. All initialization options are optional and will
    be given good defaults. You should probably at the very least give a memorable `name` and a `domain`. Variables
    can mostly be treated as strings with some extra information/utilities attached.

    With the `pyyaml` library installed, all `Variable` objects can be saved or loaded directly from a `.yml` file by
    using the `!Variable` yaml tag (which is loaded by default with `amisc`).

    - Use `Variable.distribution` to specify PDFs, such as for random variables. See the `Distribution` classes.
    - Use `Variable.norm` to specify a transformed-space that is more amenable to surrogate construction
      (e.g. mapping to the range (0,1)). See the `Transform` classes.
    - Use `Variable.compression` to specify high-dimensional, coordinate-based field quantities,
      such as from the output of many simulation software programs. See the `Compression` classes.
    - Use `Variable.category` as an additional layer for using Variable's in different ways (e.g. set a "calibration"
      category for Bayesian inference).

    !!! Example
        ```python
        # Random variable
        temp = Variable(name='T', description='Temperature', units='K', distribution='Uniform(280, 320)')
        samples = temp.sample(100)
        pdf = temp.pdf(samples)

        # Field quantity
        vel = Variable(name='u', description='Velocity', units='m/s', compression={'fields': ['ux', 'uy', 'uz']})
        vel_data = ...  # from a simulation
        reduced_vel = vel.compress(vel_data)
        ```

    !!! Warning
        Changes to collection fields (like `Variable.norm`) should completely reassign the _whole_
        collection to trigger the correct validation, rather than editing particular entries. For example, reassign
        `norm=['log', 'linear(2, 2)']` rather than editing norm via `norm.append('linear(2, 2)')`.

    :ivar name: an identifier for the variable, can compare variables directly with strings for indexing purposes
    :ivar nominal: a typical value for this variable
    :ivar description: a lengthier description of the variable
    :ivar units: assumed units for the variable (if applicable)
    :ivar category: an additional descriptor for how this variable is used, e.g. calibration, operating, design, etc.
    :ivar tex: latex format for the variable, i.e. "$x_i$"
    :ivar compression: specifies field quantities and links to relevant compression data
    :ivar distribution: a string specifier of a probability distribution function (see the `Distribution` types)
    :ivar domain: the explicit domain bounds of the variable (limits of where you expect to use it);
                  for field quantities, this is a list of domains for each latent dimension
    :ivar norm: specifier of a map to a transformed-space for surrogate construction (see the `Transform` types)
    """
    yaml_tag: ClassVar[str] = u'!Variable'
    model_config = ConfigDict(validate_assignment=True, arbitrary_types_allowed=True, validate_default=True)

    name: Optional[str] = None
    nominal: Optional[float] = None
    description: Optional[str] = None
    units: Optional[str] = None
    category: Optional[str] = None
    tex: Optional[str] = None
    compression: Optional[str | dict | Compression] = None
    distribution: Optional[str | Distribution] = None
    domain: Optional[str | tuple[float, float] | list] = None
    norm: Optional[_TransformLike] = None

    def __init__(self, /, name=None, **kwargs):
        # Try to set the variable name if instantiated as "x = Variable()"
        if name is None:
            name = _inspect_assignment('Variable')
        name = name or "X_" + "".join(random.choices(string.digits, k=3))
        super().__init__(name=name, **kwargs)

    @field_validator('tex')
    @classmethod
    def _validate_tex(cls, tex: str) -> str | None:
        if tex is None:
            return tex
        if not tex.startswith('$'):
            tex = rf'${tex}'
        if not tex[-1] == '$':
            tex = rf'{tex}$'
        return tex

    @field_validator('compression')
    @classmethod
    def _validate_compression(cls, compression: str | dict | Compression, info: ValidationInfo) -> Compression | None:
        if compression is None:
            return compression
        elif isinstance(compression, str):
            return Compression.deserialize(compression)
        elif isinstance(compression, dict):
            compression['fields'] = compression.get('fields', None) or [info.data['name']]
            return Compression.from_dict(compression)
        else:
            compression.fields = compression.fields or [info.data['name']]
            return compression

    @field_validator('distribution')
    @classmethod
    def _validate_dist(cls, dist: str | Distribution) -> Distribution | None:
        if dist is None:
            return dist
        if isinstance(dist, Distribution):
            return dist
        elif isinstance(dist, str):
            return Distribution.from_string(dist)
        else:
            raise ValueError(f'Cannot convert {dist} to a Distribution object.')

    @field_validator('domain')
    @classmethod
    def _validate_domain(cls, domain: list | tuple | str, info: ValidationInfo) -> tuple | list | None:
        """Try to extract the domain from the distribution if not provided, or convert from a string.
        Returns a list of domains for each latent dimension if this is a field quantity with compression.
        """
        if domain is None:
            if dist := info.data['distribution']:
                domain = dist.domain()
            elif compression := info.data['compression']:
                domain = compression.estimate_latent_ranges()
        elif isinstance(domain, str):
            domain = tuple(ast.literal_eval(domain.strip()))
        elif isinstance(domain, list):
            domain = [tuple(ast.literal_eval(d.strip())) if isinstance(d, str) else d for d in domain]

        if domain is None:
            return domain

        if isinstance(domain, list):
            for d in domain:
                assert isinstance(d, tuple) and len(d) == 2
                assert d[1] > d[0], 'Domain must be specified as (lower_bound, upper_bound)'
        else:
            assert isinstance(domain, tuple) and len(domain) == 2
            assert domain[1] > domain[0], 'Domain must be specified as (lower_bound, upper_bound)'

        return domain

    @field_validator('norm')
    @classmethod
    def _validate_norm(cls, norm: _TransformLike, info: ValidationInfo) -> list[Transform] | None:
        if norm is None:
            return norm
        norm = Transform.from_string(norm)

        # Set default values for minmax and zscore transforms
        domain = info.data['domain']
        normal_args = None
        if dist := info.data['distribution']:
            if isinstance(dist, Normal):
                normal_args = dist.dist_args
        for transform in norm:
            if isinstance(transform, Minmax):
                if domain and np.any(np.isnan(transform.transform_args[0:2])):
                    transform.update(lb=domain[0], ub=domain[1])
            elif isinstance(transform, Zscore):
                if normal_args and np.any(np.isnan(transform.transform_args)):
                    transform.update(mu=normal_args[0], std=normal_args[1])

        return norm

    def __getitem__(self, item):
        return getattr(self, item)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        """Allows variables to be used as keys in dictionaries and to be considered equal to their string
        representations.
        """
        return hash(self.name)

    def __eq__(self, other):
        """Consider two `Variables` equal if they share the same string name

        Also returns true when checking if this `Variable` is equal to a string by itself.
        """
        if isinstance(other, Variable):
            return self.name == other.name
        elif isinstance(other, str):
            return self.name == other
        else:
            return False

    def get_tex(self, units: bool = False, symbol: bool = True) -> str:
        """Return a raw string that is well-formatted for plotting (with latex).

        :param units: whether to include the units in the string
        :param symbol: just latex symbol if true, otherwise the full description
        :returns: the latex formatted string
        """
        s = (self.tex if symbol else self.description) or self.name
        return r'{} [{}]'.format(s, self.units or '-') if units else r'{}'.format(s)

    def get_nominal(self) -> float | list | None:
        """Return the nominal value of the variable. Defaults to the mean for a normal distribution or the
        center of the domain if `var.nominal` is not specified. Returns a list of nominal values for each latent
        dimension if this is a field quantity with compression.

        :returns: the nominal value(s)
        """
        nominal = self.nominal
        if nominal is None:
            if dist := self.distribution:
                nominal = float(dist.nominal())
            elif domain := self.get_domain():
                nominal = [np.mean(d) for d in domain] if isinstance(domain, list) else float(np.mean(domain))

        return nominal

    def get_domain(self) -> tuple | list | None:
        """Return a tuple of the defined domain of this variable. Returns a list of domains for each latent dimension
        if this is a field quantity with compression.

        :returns: the domain(s) of this variable
        """
        if self.domain is None:
            return None
        elif isinstance(self.domain, list):
            return self.domain
        elif self.compression is not None:
            # Try to infer a list of domains from compression latent size
            try:
                return [self.domain] * self.compression.latent_size()
            except Exception as e:
                raise ValueError(f'Variables with `compression` data should return a list of domains, one '
                                 f'for each latent coefficient. Could not infer domain for "{self.name}".') from e
        else:
            return self.domain

    def sample_domain(self, shape: tuple | int) -> np.ndarray:
        """Return an array of the given `shape` for uniform samples over the domain of this variable. Returns
        samples for each latent dimension if this is a field quantity with compression.

        !!! Note
            The last dim of the returned samples will be the latent space size for field quantities.

        :param shape: the shape of samples to return
        :returns: the random samples over the domain of the variable
        """
        if isinstance(shape, int):
            shape = (shape, )
        if domain := self.get_domain():
            if isinstance(domain, list):
                lb = np.atleast_1d([d[0] for d in domain])
                ub = np.atleast_1d([d[1] for d in domain])
                return np.random.rand(*shape, 1) * (ub - lb) + lb
            else:
                return np.random.rand(*shape) * (domain[1] - domain[0]) + domain[0]
        else:
            raise RuntimeError(f'Variable "{self.name}" does not have a domain specified.')

    def update_domain(self, domain: tuple[float, float] | list[tuple], override: bool = False):
        """Update the domain of this variable by taking the minimum or maximum of the new domain with the current domain
        for the lower and upper bounds, respectively. Will attempt to update the domain of each latent dimension
        if this is a field quantity with compression. If the variable has a `Uniform` distribution, this will
        update the distribution's bounds too.

        :param domain: the new domain(s) to update with
        :param override: will simply set the domain to the new values rather than update against the current domain;
                         (default `False`)
        """
        def _update_domain(domain, curr_domain):
            lb, ub = domain
            ret = (lb, ub) if override else (min(lb, curr_domain[0]) if curr_domain is not None else lb,
                                             max(ub, curr_domain[1]) if curr_domain is not None else ub)
            return tuple(map(float, ret))

        curr_domain = self.get_domain()
        if isinstance(domain, list):
            if not isinstance(curr_domain, list):
                curr_domain = [curr_domain] * len(domain)
            self.domain = [_update_domain(d, curr_domain[i]) for i, d in enumerate(domain)]
        elif isinstance(curr_domain, list):
            if not isinstance(domain, list):
                domain = [domain] * len(curr_domain)
            self.domain = [_update_domain(d, curr_domain[i]) for i, d in enumerate(domain)]
        else:
            self.domain = _update_domain(domain, curr_domain)
            if (dist := self.distribution) is not None and isinstance(dist, Uniform | LogUniform):
                dist.dist_args = self.domain  # keep Uniform dist in sync

    def pdf(self, x: np.ndarray) -> np.ndarray:
        """Compute the PDF of the Variable at the given `x` locations. Will just return one's if the variable
        does not have a distribution.

        :param x: locations to compute the PDF at
        :returns: the PDF evaluations at `x`
        """
        if dist := self.distribution:
            return dist.pdf(x)
        else:
            return np.ones(x.shape)  # No pdf if no dist is specified

    def sample(self, shape: tuple | int, nominal: float | np.ndarray = None) -> np.ndarray:
        """Draw samples from this `Variable's` distribution. Just returns the nominal value of the given shape if
        this `Variable` has no distribution.

        :param shape: the shape of the returned samples
        :param nominal: a nominal value to use if applicable (i.e. a center for relative, tolerance, or normal)
        :returns: samples from the PDF of this `Variable's` distribution
        """
        if isinstance(shape, int):
            shape = (shape, )
        if nominal is None:
            nominal = self.get_nominal()

        if dist := self.distribution:
            return dist.sample(shape, nominal)
        else:
            # Variable's with no distribution
            if nominal is None:
                raise ValueError(f'Cannot sample "{self.name}" with no dist or nominal value specified.')
            elif isinstance(nominal, list | np.ndarray):
                return np.ones(shape + (len(nominal),)) * np.atleast_1d(nominal)  # For field quantities
            else:
                return np.ones(shape) * nominal

    def normalize(self, values: ArrayLike, denorm: bool = False) -> ArrayLike | None:
        """Normalize `values` based on this `Variable's` `norm` method(s). See `Transform` for available norm methods.

        !!! Note
            If this Variable's `self.norm` was specified as a list of norm methods, then each will be applied in
            sequence in the original order (and in reverse for `denorm=True`). When `self.distribution` is involved in
            the transforms (only for `minmax` and `zscore`), the `dist_args` will get normalized too at each
            transform before applying the next transform.

        :param values: the values to normalize (array-like)
        :param denorm: whether to denormalize instead using the inverse of the original normalization method
        :returns: the normalized (or unnormalized) values
        """
        if not self.norm or values is None:
            return values
        if dist := self.distribution:
            normal_dist = isinstance(dist, Normal)
        else:
            normal_dist = False

        def _normalize_single(values, transform, inverse, domain, dist_args):
            """Do a single transform. Might need to override transform_args depending on the transform."""
            transform_args = None
            if isinstance(transform, Minmax) and domain:
                transform_args = domain + transform.transform_args[2:]  # Update minmax bounds
            elif isinstance(transform, Zscore) and dist_args:
                transform_args = dist_args                              # Update N(mu, std)

            return transform.transform(values, inverse=inverse, transform_args=transform_args)

        domain = self.get_domain() or ()
        dist_args = self.distribution.dist_args if normal_dist else []
        if isinstance(domain, list):
            domain = ()  # For field quantities, domain is not used in normalization

        if denorm:
            # First, send domain and dist_args through the forward norm list (up until the last norm)
            hyperparams = [np.hstack((domain, dist_args))]
            for i, transform in enumerate(self.norm):
                domain, dist_args = tuple(hyperparams[i][:2]), tuple(hyperparams[i][2:])
                hyperparams.append(_normalize_single(hyperparams[i], transform, False, domain, dist_args))

            # Now denormalize in reverse
            hp_idx = -2
            for transform in reversed(self.norm):
                domain, dist_args = tuple(hyperparams[hp_idx][:2]), tuple(hyperparams[hp_idx][2:])
                values = _normalize_single(values, transform, True, domain, dist_args)
                hp_idx -= 1
        else:
            # Normalize values and hyperparams through the forward norm list
            hyperparams = np.hstack((domain, dist_args))
            for transform in self.norm:
                domain, dist_args = tuple(hyperparams[:2]), tuple(hyperparams[2:])
                values = _normalize_single(values, transform, denorm, domain, dist_args)
                hyperparams = _normalize_single(hyperparams, transform, denorm, domain, dist_args)

        return values

    def denormalize(self, values):
        """Alias for `normalize(denorm=True)`. See `normalize` for more details."""
        return self.normalize(values, denorm=True)

    def compress(self, values: CompressionData, coords: np.ndarray = None,
                 reconstruct: bool = False) -> CompressionData:
        """Compress or reconstruct field quantity values using this `Variable's` compression info.

        !!! Note "Specifying compression values"
            If only one field quantity is associated with this variable, then
            specify `values` as `dict(coords=..., name=...)` for this Variable's `name`. If `coords` is not specified,
            then this assumes the locations are the same as the reconstruction data (and skips interpolation).

        !!! Info "Compression workflow"
            Generally, compression follows `interpolate -> normalize -> compress` to take raw values into the compressed
            "latent" space. The interpolation step is required to make sure `values` align with the coordinates used
            when building the compression map in the first place (such as through SVD).

        :param values: a `dict` with a key for each field qty of shape `(..., qty.shape)` and a `coords` key of shape
                      `(qty.shape, dim)` that gives the coordinates of each point. Only a single `latent` key should
                      be given instead if `reconstruct=True`.
        :param coords: the coordinates of each point in `values` if `values` did not contain a `coords` key;
                       defaults to the compression grid coordinates
        :param reconstruct: whether to reconstruct values instead of compress
        :returns: the compressed values with key `latent` and shape `(..., latent_size)`; if `reconstruct=True`,
                  then the reconstructed values with shape `(..., qty.shape)` for each `qty` key are returned.
                  The return `dict` also has a `coords` key with shape `(qty.shape, dim)`.
        """
        if not self.compression:
            raise ValueError(f'Compression is not supported for variable "{self.name}". Please specify a compression'
                             f' method for this variable.')
        if not self.compression.map_exists:
            raise ValueError(f'Compression map not computed yet for "{self.name}".')

        # Default field coordinates to the compression coordinates if they are not provided
        field_coords = values.pop('coords', coords)
        if field_coords is None:
            field_coords = self.compression.coords
        ret_dict = {'coords': field_coords}

        # For reconstruction: decompress -> denormalize -> interpolate
        if reconstruct:
            try:
                states = np.atleast_1d(values['latent'])    # (..., rank)
            except KeyError as e:
                raise ValueError('Must pass values["latent"] in for reconstruction.') from e
            states = self.compression.reconstruct(states)   # (..., dof)
            states = self.denormalize(states)               # (..., dof)
            states = self.compression.interpolate_from_grid(states, field_coords)
            ret_dict.update(states)

        # For compression: interpolate -> normalize -> compress
        else:
            states = self.compression.interpolate_to_grid(field_coords, values)
            states = self.normalize(states)                 # (..., dof)
            states = self.compression.compress(states)      # (..., rank)
            ret_dict['latent'] = states

        return ret_dict

    def reconstruct(self, values, coords=None):
        """Alias for `compress(reconstruct=True)`. See `compress` for more details."""
        return self.compress(values, coords=coords, reconstruct=True)

    def serialize(self, save_path: str | Path = '.') -> dict:
        """Convert a `Variable` to a `dict` with only standard Python types
        (i.e. convert custom objects like `dist` and `norm` to strings and save `compression` to a `.pkl`).

        :param save_path: the path to save the compression data to (defaults to current directory)
        :returns: the serialized `dict` of the `Variable` object
        """
        d = {}
        for key, value in self.__dict__.items():
            if value is not None and not key.startswith('_'):
                if key == 'domain':
                    d[key] = [str(v) for v in value] if isinstance(value, list) else str(value)
                elif key == 'distribution':
                    d[key] = str(value)
                elif key == 'norm':
                    d[key] = [str(transform) for transform in value]
                elif key == 'compression':
                    fname = f'{self.name}_compression.pkl'
                    d[key] = value.serialize(save_path=Path(save_path) / fname)
                else:
                    d[key] = value
        return d

    @classmethod
    def deserialize(cls, data: dict, search_paths: list[str | Path] = None) -> Variable:
        """Convert a `dict` to a `Variable` object. Let `pydantic` handle validation and conversion of fields.

        :param data: the `dict` to convert to a `Variable`
        :param search_paths: the paths to search for compression files (if necessary)
        :returns: the `Variable` object
        """
        if isinstance(data, Variable):
            return data
        elif isinstance(data, str):
            return cls(name=data)
        else:
            if (compression := data.get('compression', None)) is not None:
                if isinstance(compression, str):
                    data['compression'] = search_for_file(compression, search_paths=search_paths)
            return cls(**data)

    @staticmethod
    def _yaml_representer(dumper: yaml.Dumper, data: Variable) -> yaml.MappingNode:
        """Convert a single `Variable` object (`data`) to a yaml MappingNode (i.e. a `dict`)."""
        save_path, save_file = _get_yaml_path(dumper)
        return dumper.represent_mapping(Variable.yaml_tag, data.serialize(save_path=save_path))

    @staticmethod
    def _yaml_constructor(loader: yaml.Loader, node):
        """Convert the `!Variable` tag in yaml to a single `Variable` object (or a list of `Variables`)."""
        save_path, save_file = _get_yaml_path(loader)
        if isinstance(node, yaml.SequenceNode):
            return [ele if isinstance(ele, Variable) else Variable.deserialize(ele, search_paths=[save_path]) for ele in
                    loader.construct_sequence(node, deep=True)]
        elif isinstance(node, yaml.MappingNode):
            return Variable.deserialize(loader.construct_mapping(node), search_paths=[save_path])
        else:
            raise NotImplementedError(f'The "{Variable.yaml_tag}" yaml tag can only be used on a yaml sequence or '
                                      f'mapping, not a "{type(node)}".')


class VariableList(OrderedDict, Serializable):
    """Store `Variables` as `str(var) : Variable` in the order they were passed in. You can:

    - Initialize/update from a single `Variable` or a list of `Variables`
    - Get/set a `Variable` directly or by name via `my_vars[var]` or `my_vars[str(var)]` etc.
    - Retrieve the original order of insertion by `list(my_vars.items())`
    - Access/delete elements by order of insertion using integer/slice indexing (i.e. `my_vars[1:3]`)
    - Save/load from yaml file using the `!VariableList` tag
    """
    yaml_tag = '!VariableList'

    def __init__(self, data: list[Variable] | Variable | OrderedDict | dict = None, **kwargs):
        """Initialize a collection of `Variable` objects."""
        super().__init__()
        self.update(data, **kwargs)

    def __iter__(self):
        yield from self.values()

    def __eq__(self, other):
        if isinstance(other, VariableList):
            for v1, v2 in zip(self.values(), other.values()):
                if v1 != v2:
                    return False
            return True
        else:
            return False

    def append(self, data: Variable):
        self.update(data)

    def extend(self, data: list[Variable]):
        self.update(data)

    def index(self, key):
        for i, k in enumerate(self.keys()):
            if k == key:
                return i
        raise ValueError(f"'{key}' is not in list")

    def get_domains(self, norm: bool = True):
        """Get normalized variable domains (expand latent coefficient domains for field quantities). Assume a
        domain of `(0, 1)` for variables if their domain is not specified.

        :param norm: whether to normalize the domains using `Variable.norm` (useful for getting bds for surrogate);
                     latent coefficient domains do not get normalized
        :returns: a `dict` of variables to their normalized domains; field quantities return a domain for each
                  of their latent coefficients
        """
        domains = {}
        for var in self:
            var_domain = var.get_domain()
            if isinstance(var_domain, list):  # only field qtys return a list of domains, one for each latent coeff
                for i, domain in enumerate(var_domain):
                    domains[f'{var.name}{LATENT_STR_ID}{i}'] = domain
            elif var_domain is None:
                domains[var.name] = (0, 1)
            else:
                domains[var.name] = var.normalize(var_domain) if norm else var_domain
        return domains

    def get_pdfs(self, norm: bool = True):
        """Get callable pdfs for all variables (skipping field quantities for now)

        :param norm: whether values passed to the pdf functions are normalized and should be denormed first
                     before pdf evaluation (useful for surrogate construction where samples are gathered in the
                     normalized space)
        :returns: a `dict` of variables to callable pdf functions; field quantities are skipped.
        """
        def _get_pdf(var, norm):
            return lambda z: var.pdf(var.denormalize(z) if norm else z)

        pdf_fcns = {}
        for var in self:
            var_domain = var.get_domain()
            if isinstance(var_domain, list):  # only field qtys return a list of domains, one for each latent coeff
                # for i, domain in enumerate(var_domain):
                    # pdf_fcns[f'{var.name}{LATENT_STR_ID}{i}'] = var.latent_pdfs[i]  TODO: Implement latent pdfs
                pass
            else:
                pdf_fcns[var.name] = _get_pdf(var, norm)
        return pdf_fcns

    def update(self, data: list[Variable | str] | str | Variable | OrderedDict | dict = None, **kwargs):
        """Update from a list or dict of `Variable` objects, or from `key=value` pairs."""
        if data:
            if isinstance(data, OrderedDict | dict):
                for key, value in data.items():
                    self.__setitem__(key, value)
            else:
                data = [data] if not isinstance(data, list | tuple) else data
                for variable in data:
                    self.__setitem__(str(variable), variable)
        if kwargs:
            for key, value in kwargs.items():
                self.__setitem__(key, value)

    def get(self, key, default=None):
        """Make sure this passes through `__getitem__()`"""
        try:
            return self.__getitem__(key)
        except Exception:
            return default

    def __setitem__(self, key, value):
        """Only allow `str(var): Variable` items. Or normal list indexing via `my_vars[0] = var`."""
        if isinstance(key, int):
            k = list(self.keys())[key]
            self.__setitem__(k, value)
            return
        if isinstance(value, str):
            value = Variable(name=value)
        if not isinstance(key, str | Variable):
            raise TypeError(f'VariableList key "{key}" is not a Variable or string.')
        if not isinstance(value, Variable):
            raise TypeError(f'VariableList value "{value}" is not a Variable.')
        super().__setitem__(str(key), value)

    def __getitem__(self, key):
        """Allow accessing variable(s) directly via `my_vars[var]` or by index/slicing."""
        if isinstance(key, list | tuple):
            return [self.__getitem__(ele) for ele in key]
        elif isinstance(key, int | slice):
            return list(self.values())[key]
        elif isinstance(key, str | Variable):
            return super().__getitem__(str(key))
        else:
            raise TypeError(f'VariableList key "{key}" is not valid.')

    def __delitem__(self, key):
        """Allow deleting variable(s) directly or by index/slicing."""
        if isinstance(key, list | tuple):
            for ele in key:
                self.__delitem__(ele)
        elif isinstance(key, int | slice):
            ele = list(self.keys())[key]
            if isinstance(ele, list):
                for item in ele:
                    super().__delitem__(item)
            else:
                super().__delitem__(ele)
        elif isinstance(key, str | Variable):
            super().__delitem__(str(key))
        else:
            raise TypeError(f'VariableList key "{key}" is not valid.')

    def __str__(self):
        return str(list(self.values()))

    def __repr__(self):
        return self.__str__()

    def serialize(self, save_path='.') -> list[dict]:
        """Convert to a list of `dict` objects for each `Variable` in the list.

        :param save_path: the path to save the compression data to (defaults to current directory)
        """
        return [var.serialize(save_path=save_path) for var in self.values()]

    @classmethod
    def merge(cls, *variable_lists) -> VariableList:
        """Merge multiple sets of variables into a single `VariableList` object.

        !!! Note
            Variables with the same name will be merged by keeping the one with the most information provided.

        :param variable_lists: the variables/lists to merge
        :returns: the merged `VariableList` object
        """
        merged_vars = cls()

        def _get_best_variable(var1, var2):
            var1_dict = {key: value for key, value in var1.__dict__.items() if value is not None}
            var2_dict = {key: value for key, value in var2.__dict__.items() if value is not None}
            return var1 if len(var1_dict) >= len(var2_dict) else var2

        for var_list in variable_lists:
            for var in cls(var_list):
                if var.name in merged_vars:
                    merged_vars[var.name] = _get_best_variable(merged_vars[var.name], var)
                else:
                    merged_vars[var.name] = var

        return merged_vars

    @classmethod
    def deserialize(cls, data: dict | list[dict], search_paths=None) -> VariableList:
        """Convert a `dict` or list of `dict` objects to a `VariableList` object. Let `pydantic` handle validation."""
        if not isinstance(data, list):
            data = [data]
        return cls([Variable.deserialize(d, search_paths=search_paths) for d in data])

    @staticmethod
    def _yaml_representer(dumper: yaml.Dumper, data: VariableList) -> yaml.SequenceNode:
        """Convert a single `VariableList` object (`data`) to a yaml SequenceNode (i.e. a list)."""
        save_path, save_file = _get_yaml_path(dumper)
        return dumper.represent_sequence(VariableList.yaml_tag, data.serialize(save_path=save_path))

    @staticmethod
    def _yaml_constructor(loader: yaml.Loader, node):
        """Convert the `!VariableList` tag in yaml to a `VariableList` object."""
        save_path, save_file = _get_yaml_path(loader)
        if isinstance(node, yaml.SequenceNode):
            return VariableList.deserialize(loader.construct_sequence(node, deep=True), search_paths=[save_path])
        elif isinstance(node, yaml.MappingNode):
            return VariableList.deserialize(loader.construct_mapping(node), search_paths=[save_path])
        else:
            raise NotImplementedError(f'The "{VariableList.yaml_tag}" yaml tag can only be used on a yaml sequence or '
                                      f'mapping, not a "{type(node)}".')



"""Test the package utilities."""
# ruff: noqa: E741
import time

import numpy as np
from scipy.linalg import lapack

from amisc.compression import SVD
from amisc.typing import LATENT_STR_ID
from amisc.utils import (
    _inspect_function,
    constrained_lls,
    format_inputs,
    get_logger,
    parse_function_string,
    to_model_dataset,
    to_surrogate_dataset,
)
from amisc.variable import Variable, VariableList


def test_inspect_function():
    """Test inspection of a function."""
    def func1(a, b=2, **kwargs):
        y1 = a + b
        return y1

    def another_function(x, y, z):
        hello = 'world'
        _there_is = 'no spoon'
        nothing = x * y + z
        left = 'right'
        return hello, _there_is, nothing, left

    def sum_function(*args, **kwargs):
        return sum(args) + sum(kwargs.values())

    test_cases = [(func1, ['a'], ['y1']),
                  (another_function, ['x', 'y', 'z'], ['hello', '_there_is', 'nothing', 'left']),
                  (sum_function, [], [])
                  ]

    for func, expected_args, expected_outputs in test_cases:
        args, outputs = _inspect_function(func)
        assert args == expected_args
        assert outputs == expected_outputs


def test_format_inputs():
    """Test broadcasting of input arrays to the same leading dimensions"""
    cases = [{'x': np.random.rand(10, 2, 5), 'y': np.random.rand(1), 'z': np.random.rand(1, 20), 'z2': 1.0},
             {'a': np.random.rand(10, 1, 2), 'b': np.random.rand(1, 5, 1), 'c': np.random.rand(1, 1)},
             {'x1': 1.0, 'x2': [2.0], 'x3': np.random.rand(20, 1), 'x4': np.random.rand(1, 20, 1)},
             {'x': np.random.rand(15), 'y': np.random.rand(15, 5, 2), 'z': np.random.rand(15, 3, 5)},
             {'a': np.random.rand(10, 2), 'b': np.random.rand(5,), 'c': np.random.rand(6, 1)}]
    results = [(10,), (10, 5), (20,), (15,), ()]
    for case, result in zip(cases, results):
        formatted_inputs, loop_shape = format_inputs(case)
        assert loop_shape == result, f"Loop shape mismatch: {loop_shape} != {result}"
        for key, value in formatted_inputs.items():
            og_array = np.atleast_1d(case[key])
            N = np.prod(result)
            l = len(result)
            if l > 0:
                assert value.shape[0] == N, f"Loop shape mismatch for {key}: {value.shape[0]} != {N}"
                assert value.shape[1:] == og_array.shape[l:], (f"Other shape mismatch for {key}: "
                                                               f"{value.shape[1:]} != {og_array.shape[l:]}")
            else:
                assert value.shape == og_array.shape, f"Shape mismatch for {key}: {value.shape} != {og_array.shape}"


def test_parse_function_string():
    """Test parsing a function/class string"""
    cases = ["MyClass(1, 'False', True, alpha=(2,2), hello='goodbye', s='()\":\"true')",
             'myfunc(1.0, 1e6, "12", *, std="file.py")',
             "AnotherClass(42, 'test', beta=[1,2,3], gamma={'key': 'value'})",
             'anotherfunc(3.14, "hello", *, flag=True)',
             "ComplexClass(0, 'complex', delta=(1,2,3), epsilon={'nested': {'key': 'value'}})"
             ]
    results = [('MyClass', [1, 'False', True], {'alpha': (2, 2), 'hello': 'goodbye', 's': '()":"true'}),
               ('myfunc', [1.0, 1e6, '12'], {'std': 'file.py'}),
               ('AnotherClass', [42, 'test'], {'beta': [1, 2, 3], 'gamma': {'key': 'value'}}),
               ('anotherfunc', [3.14, 'hello'], {'flag': True}),
               ('ComplexClass', [0, 'complex'], {'delta': (1, 2, 3), 'epsilon': {'nested': {'key': 'value'}}})
               ]
    for case, result in zip(cases, results):
        name, args, kwargs = parse_function_string(case)
        assert name == result[0]
        assert args == result[1]
        assert kwargs == result[2]


def test_logging():
    """Test logging and plotting utils"""
    logger = get_logger('tester', stdout=True)
    logger.info('Testing logger...')


def test_lls():
    """Test constrained linear least squares routine against scipy lapack."""
    X = 100
    Y = 100
    M = 10
    N = 10
    P = 1
    tol = 1e-8

    A = np.random.rand(X, Y, M, N)
    b = np.random.rand(X, Y, M, 1)
    C = np.random.rand(X, Y, P, N)
    d = np.random.rand(X, Y, P, 1)

    # custom solver
    t1 = time.time()
    alpha = np.squeeze(constrained_lls(A, b, C, d), axis=-1)  # (*, N)
    t2 = time.time()

    # Built in scipy solver
    alpha2 = np.zeros((X, Y, N))
    t3 = time.time()
    for i in range(X):
        for j in range(Y):
            Ai = A[i, j, ...]
            bi = b[i, j, ...]
            Ci = C[i, j, ...]
            di = d[i, j, ...]
            ret = lapack.dgglse(Ai, Ci, bi, di)
            alpha2[i, j, :] = ret[3]
    t4 = time.time()

    # Results
    diff = alpha - alpha2
    assert np.max(np.abs(diff)) < tol
    print(f'Custom CLLS time: {t2-t1} s. Scipy time: {t4-t3} s.')


def test_dataset_conversion():
    """Test conversion of datasets to/from model and surrogate forms."""
    # Data matrix for field
    param = np.random.rand(50)
    svd_coords = np.linspace(0, 10, 100)
    ux = np.sin(svd_coords)[..., np.newaxis] * param
    uy = np.cos(svd_coords)[..., np.newaxis] * param
    A = np.concatenate((ux, uy), axis=0)
    compression = SVD(data_matrix=A, rank=4, fields=['ux', 'uy'], coords=svd_coords)
    latent = compression.compress(A.T)  # (num_samples, rank)
    domain = list(zip(np.min(latent, axis=0), np.max(latent, axis=0)))

    # Scalar, normalized, and field quantity variables
    scalar = Variable('scalar', distribution='U(0, 1)')
    norm = Variable('norm', distribution='LN(0, 1)', norm='log10')
    field = Variable('f', domain=domain, compression=compression)
    vlist = VariableList([scalar, norm, field])

    # Surrogate dataset
    size = (10, 3)
    s = field.sample_domain(size)
    surr_ds = {f'{field}{LATENT_STR_ID}{i}': s[..., i] for i in range(s.shape[-1])}
    surr_ds.update({var: var.normalize(var.sample(size)) for var in [scalar, norm]})

    # To model dataset (denormalize/reconstruct)
    model_ds, fc = to_model_dataset(surr_ds, vlist, del_latent=True, f_coords=svd_coords)
    assert np.allclose(model_ds['scalar'], surr_ds['scalar'])
    assert np.allclose(norm.normalize(model_ds['norm']), surr_ds['norm'])
    assert np.allclose(fc['f_coords'], svd_coords)
    assert all([var in model_ds for var in ['scalar', 'norm', 'ux', 'uy']])
    assert all([LATENT_STR_ID not in var for var in model_ds])
    assert all([model_ds[var].shape == (*size, 100) for var in ['ux', 'uy']])
    latent = field.compress({var: model_ds[var] for var in ['ux', 'uy']})['latent']
    assert all([np.allclose(latent[..., i], surr_ds[f'{field}{LATENT_STR_ID}{i}']) for i in range(latent.shape[-1])])

    # Back to surrogate dataset (normalize/compress)
    surr_ds2, surr_vars = to_surrogate_dataset(model_ds, vlist, del_fields=True, **fc)
    assert all([v in surr_vars for v in surr_ds.keys()])
    assert all([var not in surr_ds2 for var in ['ux', 'uy']])
    assert all([np.allclose(surr_ds[var], surr_ds2[var]) for var in surr_ds2])



"""Profile the prediction of a surrogate model with `py-spy`.

!!! Example
    First, install `py-spy` via `pip install py-spy`. Then, run the `generate_surrogate` function to
    generate a surrogate model. Finally, run the `predict_surrogate` function using `py-spy` externally via:

    ```shell
    py-spy record --rate 100 -o profile_predict.svg -- python profile_predict.py
    ```

    This will generate an interactive flame graph in `profile_predict.svg` to show where the most time is spent.

"""
from concurrent.futures import ProcessPoolExecutor

import numpy as np

from amisc import Component, System, Variable


def compute_output(inputs):
    # Extract inputs from the dictionary
    x1 = inputs['x1']
    x2 = inputs['x2']
    x3 = inputs['x3']
    x4 = inputs['x4']
    x5 = inputs['x5']
    x6 = inputs['x6']
    x7 = inputs['x7']
    x8 = inputs['x8']
    x9 = inputs['x9']
    x10 = inputs['x10']

    # Compute the output array y with additional three- and four-way couplings
    y = (0.5 * x1 + 0.3 * x2 ** 2 - 0.1 * np.sin(x3)) + \
        (0.4 * x4 * np.exp(x5) - 0.2 * np.log(1 + x6)) + \
        (0.1 * x7 * x8 ** 2 - 0.3 * x9 * np.sqrt(x10 + 0.1)) + \
        (0.2 * np.cos(x1 * x5) * np.tan(x2 * x3)) + \
        (0.05 * x1 * x2 * x3) + \
        (0.1 * x4 * x5 * x6) + \
        (0.05 * x7 * x8 * x9 * x10) - \
        (0.05 * x1 * x5 * x9) + \
        (0.07 * np.sin(x2 * x4 * x6)) + \
        (0.08 * np.cos(x3 * x7 * x8 * x10))

    return {'y': y}


def generate_surrogate():
    inputs = [Variable(f'x{i+1}', distribution='U(0, 1)') for i in range(10)]
    outputs = [Variable('y')]
    comp = Component(compute_output, inputs, outputs, name='my_model', data_fidelity=(2,)*len(inputs),
                     vectorized=True)
    system = System(comp, root_dir='.', name='my_system')
    system.set_logger(stdout=True)

    with ProcessPoolExecutor() as executor:
        system.fit(max_iter=100, executor=executor, plot_interval=10, num_refine=500, max_tol=-np.inf)

    return system


def predict_surrogate(save_file):
    """Run several expensive surrogate predictions (and profile externally with py-spy)."""
    surr = System.load_from_file(save_file)
    sample_nums = [100, 500, 1000, 10000]

    for num_samples in sample_nums:
        inputs = surr.sample_inputs(num_samples)
        outputs = surr.predict(inputs)  # noqa: F841


if __name__ == '__main__':
    # Run this initially to generate the surrogate
    surr = generate_surrogate()
    iter_name = f'{surr.name}_iter{surr.refine_level}'
    save_file = surr.root_dir / 'surrogates' / iter_name / f"{iter_name}.yml"
    print(f'Surrogate save file at: {save_file}')

    # Run this with py-spy afterward using the generated surrogate file
    # save_file = Path('.') / 'amisc_2024-10-30T16.29.58/surrogates/my_system_iter100/my_system_iter100.yml'
    # predict_surrogate(save_file)
